import torch as T
from lpd.metrics.metric_base import MetricBase
from lpd.metrics.binary_accuracy import BinaryAccuracy
from lpd.enums.metric_method import MetricMethod

class BinaryAccuracyWithLogits(MetricBase):
    """
        BinaryAccuracyWithLogits is basically BinaryAccuracy with threshold=0.0
        when positive examples are >= 0  and negative examples are < 0
    """
    def __init__(self, threshold=0.0):
        super(BinaryAccuracyWithLogits, self).__init__(MetricMethod.MEAN)
        self.ba = BinaryAccuracy(threshold)


    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        return self.ba(y_pred, y_true)



from tests.test_metrics import TestMetrics
from tests.test_trainer import TestTrainer
from tests.test_predictor import TestPredictor
from tests.test_callbacks import TestCallbacks
import unittest

import examples.multiple_inputs.train as multiple_inputs_example
import examples.basic.train as basic_example
import examples.scheduler_step_on_batch_end.train as scheduler_step_on_batch_end_example
import examples.data_loader.train as data_loader_example
import examples.save_and_load.train as save_and_load_example
import examples.keras_decay_scheduler.train as keras_decay_scheduler_example
import examples.accumulate_grads.train as accumulate_grads_example
import examples.train_evaluate_predict.train as train_evaluate_predict_example
import examples.predictor.train_save_load_predict as train_save_load_predict_example
import examples.confusion_matrix.train as confusion_matrix_example

# EXAMPLES
basic_example.run()
scheduler_step_on_batch_end_example.run()
multiple_inputs_example.run()
data_loader_example.run()
save_and_load_example.run()
keras_decay_scheduler_example.run()
accumulate_grads_example.run()
train_evaluate_predict_example.run()
train_save_load_predict_example.run()
confusion_matrix_example.run()
print('----------------------------------------------------------------------')
print('Ran all examples')

# TESTS
unittest.main()


from enum import Enum, auto

class State(Enum):
    """
        The state of the trainer
        TRAIN - when the epoch is in the train mode
        VAL - when the epoch is in the validation mode
        TEST - when calling evaluate(...)
        EXTERNAL - all other states, for example, after valiadaion is done and before the next epoch train has begun
    """
    EXTERNAL = auto()
    TRAIN    = auto()
    VAL      = auto() 
    TEST     = auto()
    PREDICT  = auto()
    def __str__(self):
        return self.name


from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_context import CallbackContext
from typing import Union, List, Optional, Dict, Iterable

class CallbackBase():
    """
        These are the current available phases and states

            State.EXTERNAL
            Phase.TRAIN_BEGIN
            # train loop:
                Phase.EPOCH_BEGIN

                State.TRAIN
                # batches loop:
                    Phase.BATCH_BEGIN
                    # batch
                    Phase.BATCH_END
                State.VAL
                # batches loop:
                    Phase.BATCH_BEGIN
                    # batch
                    Phase.BATCH_END
                State.EXTERNAL

                Phase.EPOCH_END
            Phase.TRAIN_END

        Evaluation phases and states will behave as follow

            State.EXTERNAL
            Phase.TEST_BEGIN
            State.TEST
            # batches loop:
                Phase.BATCH_BEGIN
                # batch
                Phase.BATCH_END
            State.EXTERNAL
            Phase.TEST_END

        Predict phases and states will behave as follow
            State.EXTERNAL
            Phase.PREDICT_BEGIN
            State.PREDICT
            # batches loop:
                Phase.BATCH_BEGIN
                # batch
                Phase.BATCH_END
            State.EXTERNAL
            Phase.PREDICT_END

        Agrs:
            apply_on_phase - (lpd.enums.Phase) the phase to invoke this callback
            apply_on_states - (lpd.enums.State) state or list of states to invoke this parameter (under the relevant phase), None will invoke it on all states
            round_values_on_print_to - optional, it will round the numerical values in the prints
    """

    def __init__(self, apply_on_phase: Phase, 
                       apply_on_states: Union[State, List[State]], 
                       round_values_on_print_to: Optional[int]=None):
        self.apply_on_phase = apply_on_phase
        self.apply_on_states = self._extract_apply_on_states(apply_on_states)
        self.round_values_on_print_to = round_values_on_print_to
        self._validations()

    def __call__(self, callback_context: CallbackContext):
        raise NotImplementedError('Implement this function')

    def _extract_apply_on_states(self, apply_on_states):
        result = set()
        if isinstance(apply_on_states, State):
            result.add(apply_on_states)
            return result
        elif isinstance(apply_on_states, Iterable):
            for s in apply_on_states:
                if isinstance(s, State) or s is None:
                    result.add(s)
                else:
                    raise ValueError(f'[CallbackBase] - {s} is of type {type(s)}, expected type {State}')
            return result
        elif apply_on_states is None:
            result.add(apply_on_states)
            return result

        raise ValueError(f'[CallbackBase] - got bad value for apply_on_states')

    def _validations(self):
        if self.apply_on_phase is None:
            raise ValueError('[CallbackBase] - No callback phase was provided')
        if None in self.apply_on_states:
            print('[CallbackBase][!] - apply_on_states is None, callback will be applied to all states')

        valid_pairs = {
                        Phase.TRAIN_BEGIN:{None, State.EXTERNAL}, 
                        Phase.TRAIN_END:{None, State.EXTERNAL}, 
                        Phase.EPOCH_BEGIN:{None, State.EXTERNAL}, 
                        Phase.EPOCH_END:{None, State.EXTERNAL}, 
                        Phase.BATCH_BEGIN:{None, State.TRAIN, State.VAL, State.TEST, State.PREDICT}, 
                        Phase.BATCH_END:{None, State.TRAIN, State.VAL, State.TEST, State.PREDICT}, 
                        Phase.TEST_BEGIN:{None, State.EXTERNAL}, 
                        Phase.TEST_END:{None, State.EXTERNAL}, 
                        Phase.PREDICT_BEGIN:{None, State.EXTERNAL}, 
                        Phase.PREDICT_END:{None, State.EXTERNAL}, 
                        }

        if self.apply_on_states is not None:
            for state in self.apply_on_states:
                if state not in valid_pairs[self.apply_on_phase]:
                    valid_print = ' or '.join([str(s) for s in (valid_pairs[self.apply_on_phase]-{None})])
                    raise ValueError(f'[CallbackBase] - State {state} cannot be applied in Phase {self.apply_on_phase}, did you mean {valid_print} ?')

    def get_description(self):
        str_apply_on_states = ','.join([str(s) for s in self.apply_on_states])
        return f'[{self.__class__.__name__}] - apply_on_phase: {self.apply_on_phase}, apply_on_states: {str_apply_on_states}'

    def round_to(self, value: Union[float,list]):
        if self.round_values_on_print_to:
            if isinstance(value, list):
                return [round(v, self.round_values_on_print_to) for v in value]
            if isinstance(value, float):
                return round(value, self.round_values_on_print_to)
        return value

    def should_apply_on_phase(self, callback_context: CallbackContext):
        if isinstance(self.apply_on_phase, Phase):
            return callback_context.trainer_phase == self.apply_on_phase
        raise ValueError('[CallbackBase] - got bad value for apply_on_phase')

    def should_apply_on_state(self, callback_context: CallbackContext):
        if None in self.apply_on_states:
            return True

        for state in self.apply_on_states:
            if callback_context.trainer_state == state:
                return True

        return False


import torch as T
from lpd.metrics.metric_base import MetricConfusionMatrixBase
from lpd.enums.confusion_matrix_based_metric import ConfusionMatrixBasedMetric as metric

class TruePositives(MetricConfusionMatrixBase):
    """
        Agrs:
            num_classes - The number of classes in the classification
            labels - names of classes, for nice prints, if not provided, the class index will be the label
            predictions_to_classes_convertor - (optional) a function that takes y_pred batch and y_true batch and converts it into class indices batch where
                                               each index represents the chosen class
                                               if None:  
                                                        torch.max with indices will be used for multi-class.
                                                        threshold will be used for binary or multilabel.
            threshold - for binary or multilable classification
    """

    def __init__(self, num_classes, labels=None, predictions_to_classes_convertor=None, threshold=0.5):
        super(TruePositives, self).__init__(num_classes, labels, predictions_to_classes_convertor, threshold)

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        result_per_class = self.get_stats(metric.TP)
        if self.is_binary(y_pred, y_true):
            return result_per_class[1]
        return result_per_class

class TrueNegatives(MetricConfusionMatrixBase):
    """
        Agrs:
            num_classes - The number of classes in the classification
            labels - names of classes, for nice prints, if not provided, the class index will be the label
            predictions_to_classes_convertor - (optional) a function that takes y_pred batch and y_true batch and converts it into class indices batch where
                                               each index represents the chosen class
                                               if None:  
                                                        torch.max with indices will be used for multi-class.
                                                        threshold will be used for binary or multilabel.
            threshold - for binary or multilable classification
    """

    def __init__(self, num_classes, labels=None, predictions_to_classes_convertor=None, threshold=0.5):
        super(TrueNegatives, self).__init__(num_classes, labels, predictions_to_classes_convertor, threshold)

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        result_per_class = self.get_stats(metric.TN)
        if self.is_binary(y_pred, y_true):
            return result_per_class[1]
        return result_per_class

class FalseNegatives(MetricConfusionMatrixBase):
    """
        Agrs:
            num_classes - The number of classes in the classification
            labels - names of classes, for nice prints, if not provided, the class index will be the label
            predictions_to_classes_convertor - (optional) a function that takes y_pred batch and y_true batch and converts it into class indices batch where
                                               each index represents the chosen class
                                               if None:  
                                                        torch.max with indices will be used for multi-class.
                                                        threshold will be used for binary or multilabel.
            threshold - for binary or multilable classification
    """

    def __init__(self, num_classes, labels=None, predictions_to_classes_convertor=None, threshold=0.5):
        super(FalseNegatives, self).__init__(num_classes, labels, predictions_to_classes_convertor, threshold)

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        result_per_class = self.get_stats(metric.FN)
        if self.is_binary(y_pred, y_true):
            return result_per_class[1]
        return result_per_class

class FalsePositives(MetricConfusionMatrixBase):
    """
        Agrs:
            num_classes - The number of classes in the classification
            labels - names of classes, for nice prints, if not provided, the class index will be the label
            predictions_to_classes_convertor - (optional) a function that takes y_pred batch and y_true batch and converts it into class indices batch where
                                               each index represents the chosen class
                                               if None:  
                                                        torch.max with indices will be used for multi-class.
                                                        threshold will be used for binary or multilabel.
            threshold - for binary or multilable classification
    """

    def __init__(self, num_classes, labels=None, predictions_to_classes_convertor=None, threshold=0.5):
        super(FalsePositives, self).__init__(num_classes, labels, predictions_to_classes_convertor, threshold)

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        result_per_class = self.get_stats(metric.FP)
        if self.is_binary(y_pred, y_true):
            return result_per_class[1]
        return result_per_class


from enum import Enum, auto

class ConfusionMatrixBasedMetric(Enum):
    TP                          = auto()
    FP                          = auto()
    FN                          = auto()
    TN                          = auto()
    PRECISION                   = auto()
    SENSITIVITY                 = auto()
    SPECIFICITY                 = auto()
    RECALL                      = auto()
    POSITIVE_PREDICTIVE_VALUE   = auto()
    NEGATIVE_PREDICTIVE_VALUE   = auto()
    ACCURACY                    = auto()
    F1SCORE                     = auto()

    def __str__(self):
        return self.name


from lpd.callbacks.loss_optimizer_handler_base import LossOptimizerHandlerBase
from lpd.enums import Phase, State
from lpd.callbacks.callback_context import CallbackContext
from typing import Union, List, Callable


class LossOptimizerHandler(LossOptimizerHandlerBase):
    """
        The basic loss and optimizer handler to invoke loss.backward(), optimizer.step() and optimizer.zero_grad()

        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            loss_handler - a Callable to handle loss.backward(...)
                           if None, will call loss.backward()
            optimizer_step_handler - a Callable to handle optimizer.step(...)
                                     if None, will call optimizer.step()
            optimizer_zero_grad_handler - a Callable to handle optimizer.zero_grad(...)
                                          if None, will call optimizer.zero_grad()
    """

    def __init__(self, 
                    apply_on_phase: Phase=Phase.BATCH_END, 
                    apply_on_states: Union[State, List[State]]=State.TRAIN,
                    loss_handler: Callable[[CallbackContext], None] = None,
                    optimizer_step_handler: Callable[[CallbackContext], None] = None,
                    optimizer_zero_grad_handler: Callable[[CallbackContext], None] = None):
        super(LossOptimizerHandler, self).__init__(apply_on_phase, apply_on_states)
        self.loss_handler = loss_handler or self._default_loss_handler
        self.optimizer_step_handler = optimizer_step_handler or self._default_optimizer_step_handler
        self.optimizer_zero_grad_handler = optimizer_zero_grad_handler or self._default_optimizer_zero_grad_handler

    def _default_loss_handler(self, callback_context: CallbackContext):
        loss = callback_context.train_last_loss
        loss.backward()

    def _default_optimizer_step_handler(self, callback_context: CallbackContext):
        optimizer = callback_context.optimizer
        optimizer.step()

    def _default_optimizer_zero_grad_handler(self, callback_context: CallbackContext):
        optimizer = callback_context.optimizer
        optimizer.zero_grad()

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD

        self.loss_handler(c)

        self.optimizer_step_handler(c)

        self.optimizer_zero_grad_handler(c)


import torch as T
import torch.nn as nn
import torch.optim as optim

from lpd.trainer import Trainer
from lpd.extensions.custom_layers import TransformerEncoderStack, Attention, MatMul2D
from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks import StatsPrint, ModelCheckPoint, Tensorboard, EarlyStopping, SchedulerStep, LossOptimizerHandler, CallbackMonitor
from lpd.metrics import BinaryAccuracyWithLogits, TruePositives
from lpd.extensions.custom_schedulers import DoNothingToLR
import lpd.utils.torch_utils as tu


class TestModel(nn.Module):
    def __init__(self, config, num_embeddings):
        super(TestModel, self).__init__()
        self.config = config
        self.num_embeddings = num_embeddings

        #LAYERS
        self.embedding_layer = nn.Embedding(num_embeddings=num_embeddings, 
                                            embedding_dim=config.EMBEDDINGS_SIZE)
        nn.init.uniform_(self.embedding_layer.weight, a=-0.05, b=0.05) # I PREFER THIS INIT

        self.transformer_encoder = TransformerEncoderStack(in_dim=config.EMBEDDINGS_SIZE, 
                                                            key_dim=config.TRANSFORMER_KEY_DIM,
                                                            out_dim=config.EMBEDDINGS_SIZE,
                                                            num_encoders=config.NUM_TRANSFORMER_ENCODERS,
                                                            num_heads=config.NUM_HEADS_PER_TRANSFORMER,
                                                            drop_out_proba=config.TRANSFORMER_DROP_OUT_PROBA,
                                                            ff_expansion_rate=config.TRANSFORMER_FF_EXPANSION_RATE)

        self.external_query_attention = Attention(key_dim=config.EMBEDDINGS_SIZE, use_query_dense=True)
        self.norm = nn.LayerNorm(normalized_shape=config.EMBEDDINGS_SIZE) # WILL APPLY NORM OVER THE LAST DIMENTION ONLY
        self.mat_mul2d = MatMul2D(transpose_b=True)

    def forward(self, x1, x2, x3):
        # x1   : sequence-Input  	(batch, num_elements)
        # x2   : some1-Input        (batch, 1)
        # x3   : some2-Input        (batch, 1)

        x1_emb = self.embedding_layer(x1)                                                     # (batch, num_elements, emb_size)
        x1_emb_transformed = self.transformer_encoder(x1_emb)                                 # (batch, num_elements, emb_size)
        
        x3_emb = self.embedding_layer(x3)                                                     # (batch, emb_size)
        x3_emb_unsqueesed = x3_emb.unsqueeze(1)                                               # (batch, 1, emb_size)

        x1_with_x3_reduced = self.external_query_attention(q=x3_emb_unsqueesed, 
                                                           k=x1_emb_transformed, 
                                                           v=x1_emb_transformed)              # (batch, 1, emb_size)
        

        x1_with_x3_residual = self.norm(x1_with_x3_reduced + x3_emb_unsqueesed)     		  # (batch, 1, emb_size)

        x2_emb = self.embedding_layer(x2)                                                     # (batch, emb_size)

        dot_product = self.mat_mul2d(x2_emb.unsqueeze(1), x1_with_x3_residual)                  # (batch, 1, 1)
        dot_product = dot_product.squeeze(2).squeeze(1)  # safe on batch_size = 1             # (batch)
        return dot_product #NOTICE! LOGITS OUT, NOT SIGMOID, THE SIGMOID WILL BE APPLIED IN THE LOSS HANDLER FOR THIS EXAMPLE

def get_trainer(config, 
                num_embeddings,                         
                train_data_loader, 
                val_data_loader,
                train_steps,
                val_steps,
                checkpoint_dir,
                checkpoint_file_name,
                summary_writer_dir):
    device = tu.get_gpu_device_if_available()

    model = TestModel(config, num_embeddings).to(device)
   
    optimizer = optim.SGD(params=model.parameters(), lr=config.LEARNING_RATE, momentum=0.9)

    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', patience=config.EARLY_STOPPING_PATIENCE // 2, verbose=True) # needs SchedulerStep callback WITH scheduler_parameters_func
    
    loss_func = nn.BCEWithLogitsLoss().to(device)

    metric_name_to_func = {
                           "Accuracy":BinaryAccuracyWithLogits(),
                           "TP":TruePositives(2, threshold = 0)
                           }

    callbacks = [   
                    LossOptimizerHandler(),
                    SchedulerStep(scheduler_parameters_func=lambda callback_context: callback_context.val_stats.get_loss()),
                    
                    ModelCheckPoint(checkpoint_dir=checkpoint_dir, 
                                    checkpoint_file_name=checkpoint_file_name, 
                                    callback_monitor=CallbackMonitor(patience=-1,
                                                                     monitor_type=MonitorType.LOSS, 
                                                                     stats_type=StatsType.VAL, 
                                                                     monitor_mode=MonitorMode.MIN),
                                    save_best_only=True, 
                                    round_values_on_print_to=7), 
                    Tensorboard(summary_writer_dir=summary_writer_dir),
                    EarlyStopping(apply_on_phase=Phase.EPOCH_END, 
                                  apply_on_states=State.EXTERNAL,
                                  callback_monitor=CallbackMonitor(patience=config.EARLY_STOPPING_PATIENCE, 
                                                                    monitor_type=MonitorType.LOSS, 
                                                                    stats_type=StatsType.VAL, 
                                                                    monitor_mode=MonitorMode.MIN)),
                    StatsPrint(apply_on_phase=Phase.EPOCH_END, round_values_on_print_to=7, print_confusion_matrix_normalized=True) # BETTER TO PUT StatsPrint LAST (MAKES BETTER SENSE IN THE LOG PRINTS)
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=train_data_loader, 
                      val_data_loader=val_data_loader,
                      train_steps=train_steps,
                      val_steps=val_steps,
                      callbacks=callbacks,
                      name='Multi-Input-Example')
    return trainer



from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_context import CallbackContext
from typing import Union, List, Optional, Dict
from math import inf
import torch as T

class CallbackMonitor():
    """
    Will check if the desired metric improved with support for patience
    Agrs:
        patience - int or None (will be set to inf) track how many epochs/iterations without improvements in monitoring
                    (negative number will set to inf)
        monitor_type - e.g lpd.enums.MonitorType.LOSS
        stats_type - e.g lpd.enums.StatsType.VAL
        monitor_mode - e.g. lpd.enums.MonitorMode.MIN, min wothh check if the metric decreased, MAX will check for increase
        metric_name - in case of monitor_mode=lpd.enums.MonitorMode.METRIC, provide metric_name, otherwise, leave it None
    """
    def __init__(self, patience: int, monitor_type: MonitorType, stats_type: StatsType, monitor_mode: MonitorMode, metric_name: Optional[str]=None):
        self.patience = inf if patience is None or patience < 0 else patience
        self.patience_countdown = self.patience
        self.monitor_type = monitor_type
        self.stats_type = stats_type
        self.monitor_mode = monitor_mode
        self.metric_name = metric_name
        self.minimum = inf
        self.maximum = -inf
        self.previous = self._get_best()
        self.description = self._get_description()
        self._track_invoked = False 

    def _get_description(self):
        desc = f'{self.monitor_mode}_{self.stats_type}_{self.monitor_type}'
        if self.metric_name:
            return desc + f'_{self.metric_name}'
        return desc

    def _get_best(self):
        return self.minimum if self.monitor_mode == MonitorMode.MIN else self.maximum

    def track(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD

        # EXTRACT value_to_consider
        if self.monitor_type == MonitorType.LOSS:

            if self.stats_type == StatsType.TRAIN:
                value_to_consider = c.train_stats.get_loss()
            elif self.stats_type == StatsType.VAL:
                value_to_consider = c.val_stats.get_loss()

        elif self.monitor_type == MonitorType.METRIC:

            if self.stats_type == StatsType.TRAIN:
                metrics_to_consider = c.train_stats.get_metrics()
            elif self.stats_type == StatsType.VAL:
                metrics_to_consider = c.val_stats.get_metrics()

            if self.metric_name not in metrics_to_consider:
                raise ValueError(f'[CallbackMonitor] - monitor_mode={MonitorType.METRIC}, but cant find metric with name {self.metric_name}')
            value_to_consider = metrics_to_consider[self.metric_name]

        if not self._track_invoked:
            self.minimum = -T.log(T.zeros_like(value_to_consider))  # [[inf,inf,inf,inf]]
            self.maximum = T.log(T.zeros_like(value_to_consider)) # [[-inf,-inf,-inf,-inf]]
            self._track_invoked = True


        # MONITOR
        self.patience_countdown = max(0, self.patience_countdown - 1)
        change_from_previous = value_to_consider - self.previous
        curr_best = self._get_best()
        change_from_best = value_to_consider - curr_best
        curr_minimum = self.minimum
        curr_maximum = self.maximum
        self.minimum = T.min(self.minimum, value_to_consider)
        self.maximum = T.max(self.maximum, value_to_consider)
        curr_previous = self.previous
        self.previous = value_to_consider
        did_improve = False
        new_best = self._get_best()
        name = self.metric_name if self.metric_name else 'loss'

        if len(value_to_consider.shape) == 0 or  \
           (len(value_to_consider.shape) == 1 and value_to_consider.shape[0] == 1):
            if  self.monitor_mode == MonitorMode.MIN and value_to_consider < curr_minimum or \
                self.monitor_mode == MonitorMode.MAX and value_to_consider > curr_maximum:
                did_improve = True
                self.patience_countdown = self.patience
        else:
            if self.patience != inf:
                raise ValueError("[CallbackMonitor] - can't monitor patience for metric that has multiple values")
        
        return CallbackMonitorResult(did_improve=did_improve, 
                                     new_value=value_to_consider, 
                                     prev_value=curr_previous,
                                     new_best=new_best,
                                     prev_best=curr_best,
                                     change_from_previous=change_from_previous,
                                     change_from_best=change_from_best,
                                     patience_left=self.patience_countdown, 
                                     description=self.description,
                                     name = name)


class CallbackMonitorResult():
    def __init__(self, did_improve: bool, 
                        new_value: float, 
                        prev_value: float,
                        new_best: float,
                        prev_best: float,
                        change_from_previous: float,
                        change_from_best: float,
                        patience_left: int,
                        description: str,
                        name: str):
        self.name = name
        self.did_improve = did_improve
        self.new_value = new_value
        self.prev_value = prev_value
        self.new_best = new_best
        self.prev_best = prev_best
        self.change_from_previous = change_from_previous
        self.change_from_best = change_from_best
        self.patience_left = patience_left
        self.description = description

    def has_improved(self):
        return self.did_improve

    def has_patience(self):
        return self.patience_left > 0



from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitorResult
from lpd.utils.torch_utils import get_lrs_from_optimizer
from typing import Union, List, Optional, Dict, Callable

class SchedulerStep(CallbackBase):
    """This callback will invoke a "step()" on the scheduler.

        Agrs:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            scheduler_parameters_func - Since some schedulers takes parameters in step(param1, param2...)
                And other schedulers step() are parameterless, provide:
                a function (or lambda) that except CallbackContext and returns whatever information needed,
                e.g. for scheduler that takes val_loss as parameter, initialize like this:
                    SchedulerStep(scheduler_parameters_func=lambda callback_context: callback_context.val_stats.get_loss())
                if your scheduler step does not expect parameters, leave scheduler_parameters_func = None
            verbose - prints visibility, 0=no prints, 1=all prints, 2=errors/warnings only
    """
    def __init__(self, apply_on_phase: Phase=Phase.EPOCH_END, 
                       apply_on_states: Union[State, List[State]]=State.EXTERNAL,
                       scheduler_parameters_func=None,
                       verbose: int=0):
        super(SchedulerStep, self).__init__(apply_on_phase=apply_on_phase, apply_on_states=apply_on_states)
        self.scheduler_parameters_func = scheduler_parameters_func
        self.verbose = verbose

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD

        if self.verbose and c.scheduler is None:
            print('[SchedulerStep] - no scheduler defined in trainer')
            return

        current_lrs = get_lrs_from_optimizer(c.optimizer)

        if self.scheduler_parameters_func:
            c.scheduler.step(self.scheduler_parameters_func(c))
        else:
            c.scheduler.step()

        new_lrs = get_lrs_from_optimizer(c.optimizer)
        
        if self.verbose == 1:
            print(f'[SchedulerStep] - current_lrs: {current_lrs} new_lrs: {new_lrs}')



import torch as T
from lpd.metrics.metric_base import MetricBase
from lpd.enums.metric_method import MetricMethod


class TopKCategoricalAccuracy(MetricBase):
    """
        Take top k predicted classes from our model and see if the correct class was selected as top k. 
        If it was we say that our model was correct.
    """
    def __init__(self, k):
        super(TopKCategoricalAccuracy, self).__init__(MetricMethod.MEAN)
        self.k = k

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        indices = T.topk(y_pred, self.k, dim=1)[1]
        expanded_y_true = y_true.view(-1, 1).expand(-1, self.k)
        correct = T.sum(T.eq(indices, expanded_y_true), dim=1)
        accuracy = correct.float().sum() / y_true.numel()
        return accuracy


import torch as T
import torch.nn as nn
import torch.optim as optim

from lpd.trainer import Trainer
from lpd.callbacks import StatsPrint, SchedulerStep, LossOptimizerHandler
from lpd.extensions.custom_schedulers import KerasDecay
from lpd.enums import Phase, State 
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu
from lpd.metrics import CategoricalAccuracyWithLogits


def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 64, 1000, 100, 10
    num_epochs = 5
    data_loader = eu.examples_data_generator(N, D_in, D_out, category_out=True)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps

def get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps):

    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.CrossEntropyLoss().to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=0.1)

    # HERE WE USE KerasDecay, IT WILL DECAY THE LEARNING-RATE USING
    # THE FORMULA USED IN KERAS:
    # LR = INIT_LR * (1./(1. + decay * step))
    # NOTICE THAT step CAN BE BATCH/EPOCH
    # WE WILL RUN IT ON EPOCH LEVEL IN THIS EXAMPLE, SO WE EXPECT WITH DECAY = 0.01:
    
    # EPOCH 0 LR: 0.1 <--- THIS IS THE STARTING POINT
    # EPOCH 1 LR: 0.1 * (1./(1. + 0.01 * 1)) = 0.09900990099
    # EPOCH 2 LR: 0.1 * (1./(1. + 0.01 * 2)) = 0.09803921568
    # EPOCH 3 LR: 0.1 * (1./(1. + 0.01 * 3)) = 0.09708737864
    # EPOCH 4 LR: 0.1 * (1./(1. + 0.01 * 4)) = 0.09615384615
    # EPOCH 5 LR: 0.1 * (1./(1. + 0.01 * 5)) = 0.09523809523
    scheduler = KerasDecay(optimizer, decay=0.01, last_step=-1) 
    
    metric_name_to_func = {'acc':CategoricalAccuracyWithLogits()}

    callbacks = [   
                    LossOptimizerHandler(),
                    SchedulerStep(apply_on_phase=Phase.EPOCH_END,
                                  apply_on_states=State.EXTERNAL,
                                  verbose=1),                        #LET'S PRINT TO SEE THE ACTUAL CHANGES
                    StatsPrint()
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader, 
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Keras-Decay-Example')
    return trainer

def run():
    gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

    N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps = get_parameters()

    trainer = get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps)
    
    trainer.train(num_epochs)



import torch as T

def seed_torch(seed):
    T.manual_seed(seed)
    T.cuda.manual_seed_all(seed)
    T.backends.cudnn.deterministic = True

def is_gpu_available(verbose=1):
    res = T.cuda.is_available()
    if verbose: print(f"GPU Availability: {res}")
    return res

def how_many_gpus_available():
    amount = T.cuda.device_count()
    print('GPUs amount: ', amount)
    return amount

def get_gpu_device_if_available():
    if is_gpu_available():
        device = T.device('cuda:0')
    else:
        device = T.device('cpu')
    return device

def what_torch_version_is_currently_running():
    v = T.__version__
    print(f'running torch {v}')
    return v

def save_checkpoint(dir_path, file_name, trainer, msg='', verbose=1):
    full_path = dir_path +file_name
    if verbose:
        print(f'{msg} - Saving checkpoint to {full_path}')
    checkpoint = {
                  'model': trainer.model.state_dict(),
                  'optimizer': trainer.optimizer.state_dict(),
                  'scheduler': trainer.scheduler.state_dict() if trainer.scheduler else None,
                  'epoch':trainer.epoch
                  }
    T.save(checkpoint, full_path)

def load_checkpoint(checkpoint_filepath, model, optimizer, scheduler, device, verbose=1):
    if verbose:
        print('Loading checkpoint')
    checkpoint = T.load(checkpoint_filepath, map_location=device)
    model.load_state_dict(checkpoint['model'])
    optimizer.load_state_dict(checkpoint['optimizer'])
    scheduler.load_state_dict(checkpoint['scheduler'])
    return checkpoint['epoch']

def get_lrs_from_optimizer(optimizer):
    return [group['lr'] for group in optimizer.param_groups]

def copy_model_weights(source_model, target_model):
    self.target_model.load_state_dict(self.source_model.state_dict())


from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitor, CallbackMonitorResult
from typing import Union, List, Optional, Dict

class EarlyStopping(CallbackBase):
    """
        Stop training when a monitored loss/metric has stopped improving.
        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            callback_monitor - to monitor the loss/metric and stop training when stopped improving
            verbose - 0 = no print, 1 = print all, 2 = print save only
    """

    def __init__(self, 
                    apply_on_phase: Phase=Phase.EPOCH_END, 
                    apply_on_states: Union[State, List[State]]=State.EXTERNAL,
                    callback_monitor: CallbackMonitor=None,
                    verbose: int=1):
        super(EarlyStopping, self).__init__(apply_on_phase, apply_on_states)
        if callback_monitor is None:
            raise ValueError("[EarlyStopping] - callback_monitor was not provided")
        self.monitor = callback_monitor
        self.verbose = verbose

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD

        monitor_result = self.monitor.track(c)

        if monitor_result.has_patience() and self.verbose:
            print(f'[EarlyStopping] - patience:{monitor_result.patience_left} epochs')
        
        if not monitor_result.has_patience():
            c.trainer.stop()
            if self.verbose > 0:
                print(f'[EarlyStopping] - stopping on epoch {c.epoch}')





import os
import unittest

import torch.optim as optim
import torch.nn as nn
from lpd.trainer import Trainer
from lpd.predictor import Predictor
from lpd.callbacks import LossOptimizerHandler, ModelCheckPoint, CallbackMonitor
from lpd.metrics import BinaryAccuracyWithLogits
from lpd.enums import MonitorType, MonitorMode, StatsType
import lpd.utils.torch_utils as tu
import examples.utils as eu


class TestPredictor(unittest.TestCase):
    def test_save_and_predict(self):

        save_to_dir = os.path.dirname(__file__) + '/trainer_checkpoint/'
        checkpoint_file_name = 'checkpoint'
        trainer_file_name = 'trainer'

        device = tu.get_gpu_device_if_available()

        model = eu.get_basic_model(10, 10, 10).to(device)

        loss_func = nn.BCEWithLogitsLoss().to(device)
    
        optimizer = optim.Adam(model.parameters(), lr=1e-4)

        scheduler = None
        
        metric_name_to_func = {"acc":BinaryAccuracyWithLogits()}

        callbacks = [   
                        LossOptimizerHandler(),
                        ModelCheckPoint(checkpoint_dir=save_to_dir, 
                                        checkpoint_file_name=checkpoint_file_name, 
                                        callback_monitor=CallbackMonitor(patience=-1,
                                                                        monitor_type=MonitorType.LOSS, 
                                                                        stats_type=StatsType.VAL, 
                                                                        monitor_mode=MonitorMode.MIN),
                                        save_best_only=True, 
                                        save_full_trainer=False), 
                    ]

        data_loader = eu.examples_data_generator(10, 10, 10)
        data_loader_steps = 100
        num_epochs = 5

        trainer = Trainer(model=model, 
                          device=device, 
                          loss_func=loss_func, 
                          optimizer=optimizer,
                          scheduler=scheduler,
                          metric_name_to_func=metric_name_to_func, 
                          train_data_loader=data_loader, 
                          val_data_loader=data_loader,
                          train_steps=data_loader_steps,
                          val_steps=data_loader_steps,
                          callbacks=callbacks,
                          name='Predictor-Trainer-Test')

        data_generator_for_predictions = eu.examples_prediction_data_generator(data_loader, data_loader_steps)
        sample = next(data_generator_for_predictions)[0]

        # PREDICT BEFORE TRAIN
        sample_prediction_before_train = trainer.predict_sample(sample)

        trainer.train(num_epochs, verbose=0)
        
        # PREDICT AFTER TRAIN
        sample_prediction_from_trainer = trainer.predict_sample(sample)

        # SAVE THE TRAINER
        trainer.save_trainer(save_to_dir, trainer_file_name)

        #-----------------------------------------------#
        # CREATE PREDICTOR FROM CURRENT TRAINER
        #-----------------------------------------------#
        predictor_from_trainer = Predictor.from_trainer(trainer)

        # PREDICT FROM PREDICTOR
        sample_prediction_from_predictor = predictor_from_trainer.predict_sample(sample)

        self.assertFalse((sample_prediction_before_train==sample_prediction_from_trainer).all())
        self.assertTrue((sample_prediction_from_predictor==sample_prediction_from_trainer).all())



        #-----------------------------------------------#
        # LOAD MODEL CHECKPOINT AS NEW PREDICTOR
        #-----------------------------------------------#
        fresh_device = tu.get_gpu_device_if_available()
        fresh_model = eu.get_basic_model(10, 10, 10).to(fresh_device)
        loaded_predictor = Predictor.from_checkpoint(save_to_dir,
                                                     checkpoint_file_name+'_best_only',
                                                     fresh_model, 
                                                     fresh_device)

        # PREDICT AFTER LOAD
        sample_prediction_from_loaded_predictor = loaded_predictor.predict_sample(sample)

        self.assertFalse((sample_prediction_before_train==sample_prediction_from_trainer).all())
        self.assertTrue((sample_prediction_from_loaded_predictor==sample_prediction_from_trainer).all())



        #-----------------------------------------------#
        # LOAD TRAINER CHECKPOINT AS NEW PREDICTOR
        #-----------------------------------------------#
        fresh_device = tu.get_gpu_device_if_available()
        fresh_model = eu.get_basic_model(10, 10, 10).to(fresh_device)
        loaded_predictor = Predictor.from_checkpoint(save_to_dir,
                                                     trainer_file_name,
                                                     fresh_model, 
                                                     fresh_device)

        # PREDICT AFTER LOAD
        sample_prediction_from_loaded_predictor = loaded_predictor.predict_sample(sample)

        self.assertFalse((sample_prediction_before_train==sample_prediction_from_trainer).all())
        self.assertTrue((sample_prediction_from_loaded_predictor==sample_prediction_from_trainer).all())



from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_monitor import CallbackMonitor, CallbackMonitorResult
from lpd.callbacks.stats_print import StatsPrint
from lpd.callbacks.model_checkpoint import ModelCheckPoint
from lpd.callbacks.tensorboard import Tensorboard
from lpd.callbacks.early_stopping import EarlyStopping
from lpd.callbacks.scheduler_step import SchedulerStep
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.collect_outputs import CollectOutputs
from lpd.callbacks.loss_optimizer_handler import LossOptimizerHandler
from lpd.callbacks.loss_optimizer_handler_base import LossOptimizerHandlerBase



import torch as T
from lpd.metrics.metric_base import MetricBase


class ConfusionMatrix(MetricBase):
    def __init__(self, threshold=0.5):
        self.tp = 0
        self.tn = 0
        self.fp = 0
        self.fp = 0

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        assert y_true.size() == y_pred.size()
        pred = y_pred.round().long()
        accuracy = pred.eq(y_true).float().sum() / y_true.numel()
        return accuracy





import torch as T
from tqdm import tqdm
from lpd.metrics import MetricBase
from lpd.callbacks import CallbackContext, CollectOutputs, LossOptimizerHandlerBase
from lpd.enums import State, Phase
from lpd.trainer_stats import TrainerStats, StatsResult
import lpd.utils.file_utils as fu
from lpd.extensions.custom_schedulers import DoNothingToLR

class Trainer():
    """
        class that maintains all the participating objects and stats during training and evaluation

        Args:
            model - your model (nn.Module)
            device - the device to send the inputs/labels to
            loss_func - the model's loss function
            optimizer - the model's optimizer
            scheduler - the model's scheduler (make sure you add SchedulerStep to your callbacks),
                        pass None if you dont need scheduler
            metric_name_to_func - a dictionary with string as key and metric function as value
                        e.g.   {"binary_accuracy":lpd.metrics.BinaryAccuracy()}
            train_data_loader - an iterable or generator to get the next train data batch
            val_data_loader - an iterable or generator to get the next val data batch
            train_steps - total number of steps (batches) before declaring the epoch as finished
            val_steps - total number of steps (batches) before declaring the epoch as finished
            callbacks - list of lpd.callbacks to apply during the differrent training phases
                        callbacks will be executed by the order of the list 
            name - a friendly identifier

        Methods:
            save_trainer - saving the full trainer state to a file
            load_trainer - for creating a new Trainer instance from a saved Trainer checkpoint
            summary - will print information about the trainer and the model
            stop - will indicate this trainer to stop, e.g. from a callback
            train - this is the training loop, it will invoke the training and validation phases, as well as callbacks and maintain stats
            evaluate - will run a forward pass on the test data
            predict_sample - make prediction on single sample
            predict_batch - make prediction on single batch
            predict_data_loader - make prediction on data loader (DataLoader/Iterable/Generator)
    """

    def __init__(self, model,
                       device,
                       loss_func,
                       optimizer,
                       scheduler,
                       metric_name_to_func,
                       train_data_loader,
                       val_data_loader,
                       train_steps,
                       val_steps,
                       callbacks = None,
                       name = 'lpd'):
        self.device = device
        self.model = model
        self.loss_func = loss_func
        self.optimizer = optimizer
        self.scheduler = scheduler if scheduler else DoNothingToLR()
        self.metric_name_to_func = metric_name_to_func if metric_name_to_func else {}
        self._validate_metric_name_to_func()
        self.train_data_loader = train_data_loader
        self.val_data_loader = val_data_loader
        self.train_steps = train_steps
        self.val_steps = val_steps
        self.callbacks = callbacks if callbacks else []
        self.name = name

        self.epoch = 0
        self.sample_count = 0
        self.sample_count_in_epoch = 0
        self.iteration = 0
        self.iteration_in_epoch = 0

        self.state = State.EXTERNAL
        self.phase = Phase.IDLE
        self.train_stats = TrainerStats(self.metric_name_to_func)
        self.train_last_loss = None
        self.val_stats = TrainerStats(self.metric_name_to_func)
        self.val_last_loss = None
        self.test_stats = TrainerStats(self.metric_name_to_func)
        self.test_last_loss = None

        self._stopped = False
        self._last_outputs = None
        self._last_labels = None
        self._total_num_epochs = 0

    def _validate_metric_name_to_func(self):
        for name,func in self.metric_name_to_func.items():
            if not isinstance(func, MetricBase):
                raise ValueError(f'[Trainer] - metric "{name}" is of type {type(func)}, expected {MetricBase}')

    def _train_callbacks_validation(self):
        has_loss_optimizer_handler = False
        for cb in self.callbacks:
            if isinstance(cb, LossOptimizerHandlerBase):
                has_loss_optimizer_handler = True

        if not has_loss_optimizer_handler:
            raise ValueError('[Trainer][train] - callbacks not containing LossOptimizerHandlerBase, either use LossOptimizerHandler callback, or create your own callback derived from LossOptimizerHandlerBase')

    def _train_handler(self, loss, batch_size):
        self.sample_count += batch_size
        self.sample_count_in_epoch += batch_size
        self.iteration += 1
        self.iteration_in_epoch += 1
        self.train_last_loss = loss

    def _val_handler(self, loss, batch_size):
        self.val_last_loss = loss

    def _test_handler(self, loss, batch_size):
        self.test_last_loss = loss

    def _predict_handler(self, loss, batch_size):
        pass

    def _labels_handler(self, labels):
        return labels.to(self.device)

    def _inputs_handler(self, inputs):
        if isinstance(inputs, list):
            # MULTIPLE INPUTS CONSTRUCTED IN A LIST
            return [x.to(self.device) for x in inputs]
        #SINGLE INPUT
        return [inputs.to(self.device)]

    def _get_epoch_description(self):
        if self.state == State.TEST or self.state == State.PREDICT:
            desc = f'[{self.state}]'
        elif self.state == State.VAL:
            desc = f'[Val   epoch {self.epoch}/{self._total_num_epochs}]'
        else: #TRAIN
            desc = f'[Train epoch {self.epoch}/{self._total_num_epochs}]'
        return desc

    def _tqdm_description(self, loop, stats):
        desc = self._get_epoch_description()
        loop.set_description(desc)
        if self.state != State.PREDICT:
            stats_result = StatsResult('', stats)
            if stats_result.metrics:
                loop.set_postfix(loss=stats_result.loss, metrics=stats_result.metrics)
            else:
                loop.set_postfix(loss=stats_result.loss)
            
    def _print_verbos_2(self, stats, verbose):
        if verbose != 2:
            return
        desc = self._get_epoch_description()
        stats_result = StatsResult('', stats)
        if stats_result.metrics:
            metrics_str = f', metrics={stats_result.metrics}'
        else:
            metrics_str = ''
        print(f'{desc}, loss={stats_result.loss}{stats_result.metrics_str}')

    def _prepare_next_batch(self, batch):
        if self.state == State.PREDICT:
            inputs,labels = batch, T.zeros(len(batch)) #FAKE LABELS FOR CODE CONSISTENCY, NO ACTUAL USE TO THEM 
        else:
            inputs,labels = batch
        return inputs,labels

    def _fwd_pass_base(self, data_loader, steps, state_handler, stats, loss_f, verbose):
        stats.reset()
        loop = tqdm(data_loader, total=steps-1, disable=verbose-1) #verbose-1 maps 0,2=>True, 1=>False
        self.sample_count_in_epoch = 0
        self.iteration_in_epoch = 0
        for batch in loop:
            inputs,labels = self._prepare_next_batch(batch)
            steps -= 1

            self.phase = Phase.BATCH_BEGIN
            self._invoke_callbacks()

            x = self._inputs_handler(inputs)
            y = self._labels_handler(labels)
            batch_size = len(y)
            outputs = self.model(*x)
            self._last_outputs = outputs
            self._last_labels = y
            loss = loss_f(outputs, y)
            stats.add_loss(loss, batch_size)
            stats.add_metrics(outputs, y, batch_size)
            state_handler(loss, batch_size)

            self.phase = Phase.BATCH_END
            self._invoke_callbacks()

            self._tqdm_description(loop, stats)

            if self._stopped:
                break

            if steps == 0:
                break

    def _fwd_pass_predict(self, predict_data_loader, predict_steps):
        if self._stopped:
            return
            
        with T.no_grad():
            self.model.eval()  #MARK STATUS AS EVAL
            self._fwd_pass_base(predict_data_loader, 
                                predict_steps, 
                                self._predict_handler, 
                                stats=TrainerStats({}), # NO STATS
                                loss_f=lambda outputs, y: T.Tensor([0]), # DO NOTHING LOSS
                                verbose=0) 

    def _fwd_pass_test(self, test_data_loader, test_steps, verbose):
        if self._stopped:
            return
            
        with T.no_grad():
            self.model.eval()  #MARK STATUS AS EVAL
            self._fwd_pass_base(test_data_loader, 
                                test_steps, 
                                self._test_handler, 
                                self.test_stats, 
                                loss_f=self.loss_func,
                                verbose=verbose)

    def _fwd_pass_val(self, verbose):
        if self._stopped or self.val_data_loader is None or self.val_steps == 0:
            return

        with T.no_grad():
            self.model.eval()  #MARK STATUS AS EVAL
            self._fwd_pass_base(self.val_data_loader, 
                                self.val_steps, 
                                self._val_handler, 
                                self.val_stats, 
                                loss_f=self.loss_func,
                                verbose=verbose)

    def _fwd_pass_train(self, verbose):
        if self._stopped:
            return
            
        self.model.train() #MARK STATUS AS TRAIN
        self._fwd_pass_base(self.train_data_loader, 
                            self.train_steps, 
                            self._train_handler, 
                            self.train_stats, 
                            loss_f=self.loss_func,
                            verbose=verbose)

        self._print_verbos_2(self.train_stats, verbose)

    def _invoke_callbacks(self):
        if self._stopped:
            return
        context = CallbackContext(self)
        for callback in self.callbacks:
            if callback.should_apply_on_phase(context) and \
               callback.should_apply_on_state(context):
                callback(context)

    def _predict(self, inputs_data_loader, steps):
        """
            return numpy array(s) of current trainer model's predictions.
        """

        # ADD COLLECT OUTPUTS CALLBACK 
        collect_outputs = CollectOutputs(apply_on_phase=Phase.BATCH_END, apply_on_states=State.PREDICT)
        self.callbacks.append(collect_outputs)

        self._stopped = False
        self.phase = Phase.PREDICT_BEGIN
        self._invoke_callbacks()
        self.state = State.PREDICT
        self._fwd_pass_predict(inputs_data_loader, steps)
        self.state = State.EXTERNAL
        self.phase = Phase.PREDICT_END
        self._invoke_callbacks()
        self.phase = Phase.IDLE

        # REMOVE COLLECT OUTPUTS CALLBACK 
        self.callbacks.pop()
        
        outputs = collect_outputs.get_outputs_for_state(State.PREDICT)
        return outputs


    # ---------------- PUBLIC ---------------- 


    def save_trainer(self, dir_path, file_name, msg='', verbose=1):
        full_path = dir_path + file_name
        if verbose:
            if msg:
                print(f'[Trainer {self.name}] - {msg}')
            else:
                print(f'[Trainer {self.name}] - Saving to {full_path}')

        if not fu.is_folder_exists(dir_path):
            fu.create_folder(dir_path)

        trainer_checkpoint = {
                            'model': self.model.state_dict(),
                            'loss_func': self.loss_func.state_dict(),
                            'optimizer': self.optimizer.state_dict(),
                            'scheduler':  self.scheduler.state_dict() if self.scheduler else None,
                            'metric_name_to_func': self.metric_name_to_func,
                            'callbacks': self.callbacks,
                            'name': self.name,
                            'epoch': self.epoch,
                            'iteration': self.iteration,
                            'sample_count': self.sample_count,
                            'train_stats': self.train_stats,
                            'val_stats': self.val_stats,
                            'test_stats': self.test_stats
                            }
        T.save(trainer_checkpoint, full_path)

    @staticmethod
    def load_trainer(dir_path,
                     file_name,
                     model, 
                     device,
                     loss_func, 
                     optimizer, 
                     scheduler, 
                     train_data_loader,
                     val_data_loader,
                     train_steps,
                     val_steps):
        full_path = dir_path + file_name
        checkpoint = T.load(full_path, map_location=device)
        print(f'[Trainer] - Loading from {full_path}')
        model.load_state_dict(checkpoint['model'])
        loss_func.load_state_dict(checkpoint['loss_func'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        scheduler = scheduler if scheduler else DoNothingToLR()
        scheduler.load_state_dict(checkpoint['scheduler'])

        trainer = Trainer(model=model, 
                       device=device, 
                       loss_func=loss_func, 
                       optimizer=optimizer,
                       scheduler=scheduler,
                       metric_name_to_func=checkpoint['metric_name_to_func'], 
                       train_data_loader=train_data_loader, 
                       val_data_loader=val_data_loader,
                       train_steps=train_steps,
                       val_steps=val_steps,
                       callbacks=checkpoint['callbacks'],
                       name=checkpoint['name'])
        
        if 'epoch' in checkpoint:
            trainer.epoch = checkpoint['epoch']
        if 'iteration' in checkpoint:
            trainer.iteration = checkpoint['iteration']
        if 'sample_count' in checkpoint:
            trainer.sample_count = checkpoint['sample_count']
        if 'train_stats' in checkpoint:
            trainer.train_stats = checkpoint['train_stats']
        if 'val_stats' in checkpoint:
            trainer.val_stats = checkpoint['val_stats']
        if 'test_stats' in checkpoint:
            trainer.test_stats = checkpoint['test_stats'] 
        
        return trainer

    def summary(self):
        print(f'Trainer - {self.name}')

        print('defined callbacks:')
        for c in self.callbacks:
            print(c.get_description())
        print('')

        print("parameters name and device:")
        for p in self.model.named_parameters():
            print(f'name: {p[0]}, device: {p[1].device}')
            # print(p[1].data)

        print('')
        print('optimizer', type(self.optimizer))
        print('')
        
        print('Model Summary - ')
        print(self.model)
        print('')
        
        total_params = sum(p.numel() for p in self.model.parameters())
        total_params_requires_grad = sum(p.numel() for p in self.model.parameters() if p.requires_grad)

        print(f'Total params: {total_params}')
        print(f'Trainable params: {total_params_requires_grad}')
        print(f'Non-trainable params: {total_params_requires_grad-total_params}')

    def stop(self):
        self._stopped = True

    def train(self, num_epochs, verbose=1):
        """
            num_epochs - amount of epochs to run
            verbose: 0 = no progress bar, 1 = progress bar, 2 = one line per epoch
        """
        self._total_num_epochs = self.epoch + num_epochs
        self._train_callbacks_validation()
        self._stopped = False
        self.state = State.EXTERNAL
        self.phase = Phase.TRAIN_BEGIN
        self._invoke_callbacks()

        for _ in range(num_epochs):
            self.epoch += 1

            self.phase = Phase.EPOCH_BEGIN
            self._invoke_callbacks()

            self.state = State.TRAIN
            self._fwd_pass_train(verbose)
            self.state = State.VAL
            self._fwd_pass_val(verbose)
            self.state = State.EXTERNAL

            self.phase = Phase.EPOCH_END
            self._invoke_callbacks()

            if self._stopped:
                break

        self.phase = Phase.TRAIN_END
        self._invoke_callbacks()
        self.phase = Phase.IDLE

    def evaluate(self, test_data_loader, test_steps, verbose=1):
        """
            verbose: 0 = no progress bar, 1 = progress bar
        """
        self._stopped = False
        self.phase = Phase.TEST_BEGIN
        self._invoke_callbacks()
        self.state = State.TEST
        self._fwd_pass_test(test_data_loader, test_steps, verbose)
        self.state = State.EXTERNAL
        self.phase = Phase.TEST_END
        self._invoke_callbacks()
        self.phase = Phase.IDLE

        return StatsResult(self.name, self.test_stats)

    def predict_sample(self, inputs):
        # MAKE BATCH WITH 1 SAMPLE USING unsqueeze
        outputs = self.predict_batch(inputs.unsqueeze(0))
        return outputs[0]

    def predict_batch(self, inputs):
        # MAKE ITERATOR WITH 1 BATCH 1 STEP
        outputs = self.predict_data_loader([inputs], 1)
        return outputs[0]

    def predict_data_loader(self, inputs_data_loader, steps):
        return self._predict(inputs_data_loader, steps)






import torch as T
from lpd.metrics.metric_base import MetricBase
from lpd.enums.metric_method import MetricMethod


class CategoricalAccuracy(MetricBase):
    def __init__(self):
        super(CategoricalAccuracy, self).__init__(MetricMethod.MEAN)

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        indices = T.max(y_pred, 1)[1]
        correct = T.eq(indices, y_true).view(-1)
        accuracy = correct.float().sum() / correct.shape[0]
        return accuracy


from enum import Enum, auto

class Phase(Enum):
    IDLE          = auto()
    TRAIN_BEGIN   = auto()
    TRAIN_END     = auto()
    EPOCH_BEGIN   = auto()
    EPOCH_END     = auto()
    BATCH_BEGIN   = auto()
    BATCH_END     = auto()
    TEST_BEGIN    = auto()
    TEST_END      = auto()
    PREDICT_BEGIN = auto()
    PREDICT_END   = auto()

    def __str__(self):
        return self.name


from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitor, CallbackMonitorResult
from lpd.utils.torch_utils import get_lrs_from_optimizer
from typing import Union, List, Optional, Dict, Iterable

class StatsPrint(CallbackBase):
    """
        Prints informative summary of the trainer stats including loss and metrics.
        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            round_values_on_print_to - see in CallbackBase

            train_metrics_monitors - CallbackMonitor instance or list of CallbackMonitor instances that will track the improvement of the trainer's metrics.
                                    if empty list, no monitoring will be applied to metrics.
                                    if None, it will assign CallbackMonitor with MonitorMode.MAX per each metric

                                    val_metric_monitors will be applied automatically based in train_metrics_monitors

            print_confusion_matrix - If one of the metrics is of type MetricConfusionMatrixBase, setting this to True will print the confusion matrix
                                     If False, confusion matrix will not be printed (but the metric stats will remain)
            print_confusion_matrix_normalized - same as 'print_confusion_matrix', except it prints the normalized confusion matrix
    """

    def __init__(self, apply_on_phase: Phase=Phase.EPOCH_END, 
                       apply_on_states: Union[State, List[State]]=State.EXTERNAL, 
                       round_values_on_print_to: Optional[int]=None,
                       train_metrics_monitors: Union[CallbackMonitor,Iterable[CallbackMonitor]]=None,
                       print_confusion_matrix: bool=False,
                       print_confusion_matrix_normalized: bool=False):
        super(StatsPrint, self).__init__(apply_on_phase, apply_on_states, round_values_on_print_to)
        self.train_loss_monitor = CallbackMonitor(None, MonitorType.LOSS, StatsType.TRAIN, MonitorMode.MIN)
        self.val_loss_monitor = CallbackMonitor(None, MonitorType.LOSS, StatsType.VAL, MonitorMode.MIN)
        self.train_metrics_monitors = self._parse_train_metrics_monitors(train_metrics_monitors)
        self.print_confusion_matrix = print_confusion_matrix
        self.print_confusion_matrix_normalized = print_confusion_matrix_normalized
        self.val_metric_monitors = None
        self._validate_stats_type()
        self.GREEN_PRINT_COLOR = "\033[92m"
        self.END_PRINT_COLOR = "\033[0m"

    def _parse_train_metrics_monitors(self, train_metrics_monitors):
        if isinstance(train_metrics_monitors, list):
            return train_metrics_monitors
        if isinstance(train_metrics_monitors, CallbackMonitor):
            return [train_metrics_monitors]

    def _validate_stats_type(self):
        if self.train_metrics_monitors is not None:
            for m in self.train_metrics_monitors:
                if m.stats_type != StatsType.TRAIN:
                    raise ValueError(f'[StatsPrint] - train_metrics_monitors contains monitor with stats_type {m.stats_type}, expected {StatsType.TRAIN}')
            
    def _ensure_metrics_created(self, callback_context: CallbackContext):
        if self.train_metrics_monitors is not None and self.val_metric_monitors is not None:
            # ENSURED ALREADY
            return

        metric_names = callback_context.trainer.metric_name_to_func.keys()
        if self.train_metrics_monitors is None:
            self.train_metrics_monitors = [CallbackMonitor(None, MonitorType.METRIC, StatsType.TRAIN, MonitorMode.MAX, metric_name) for metric_name in metric_names]

        self.val_metric_monitors = []
        for m in self.train_metrics_monitors:
            self.val_metric_monitors.append(CallbackMonitor(m.patience, m.monitor_type, StatsType.VAL, m.monitor_mode, m.metric_name))

    def _get_print_from_monitor_result(self, monitor_result: CallbackMonitorResult) -> str:
        r = self.round_to #READABILITY
        mtr = monitor_result #READABILITY
        return f'curr:{r(mtr.new_value)}, prev:{r(mtr.prev_value)}, best:{r(mtr.new_best)}, change_from_prev:{r(mtr.change_from_previous)}, change_from_best:{r(mtr.change_from_best)}'

    def _get_print_from_metrics(self, train_metric_monitor_results: Iterable[CallbackMonitorResult], prefix: str='') -> str:
        gdic = self._get_did_improved_colored #READABILITY 

        if len(train_metric_monitor_results) == 0:
            return 'no metrics found'
        prints = []
        pre = ''
        for monitor_result in train_metric_monitor_results:
            name = monitor_result.name
            prints.append(f'{pre}name: "{name}"{gdic(monitor_result)}, {self._get_print_from_monitor_result(monitor_result)}')
            pre = prefix # APPEND PREFIX FROM THE SECOND METRIC AND ON

        return '\n'.join(prints)

    def _get_did_improved_colored(self, monitor_result):
        if monitor_result.has_improved():
            return self.GREEN_PRINT_COLOR + ' IMPROVED' + self.END_PRINT_COLOR
        return ''

    def _get_print_confusion_matrix(self, state: State, callback_context: CallbackContext, prefix: str='') -> str:
        if not self.print_confusion_matrix and not self.print_confusion_matrix_normalized:
            if state == State.TRAIN:
                return '|   |'
            elif state == State.VAL:
                return '|'

        if state == State.TRAIN:
            total_end_delimeter = '\n|   |           |' + '_' * 20
            row_start = '|   |           |-- '
            stats = callback_context.train_stats
        elif state == State.VAL:
            total_end_delimeter = '\n|               |' + '_' * 20
            row_start = '|               |-- '
            stats = callback_context.val_stats
        
        cm = stats.confusion_matrix
        if cm is None:
            raise ValueError('[StatsPrint] - print_confusion_matrix is set to True, but no confusion matrix based metric was set on trainer')
        return f'{row_start}{cm.confusion_matrix_string(prefix, normalized = self.print_confusion_matrix_normalized)}{total_end_delimeter}'
        

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY 
        r = self.round_to #READABILITY
        gdic = self._get_did_improved_colored #READABILITY 
        gpfm = self._get_print_from_metrics #READABILITY 
        gpcm = self._get_print_confusion_matrix #READABILITY 

        self._ensure_metrics_created(c)

        # INVOKE MONITORS
        t_loss_monitor_result = self.train_loss_monitor.track(c)
        v_loss_monitor_result = self.val_loss_monitor.track(c)

        train_metric_monitor_results = [m.track(c) for m in self.train_metrics_monitors]
        val_metric_monitor_results = [m.track(c) for m in self.val_metric_monitors]

        current_lrs = get_lrs_from_optimizer(c.trainer.optimizer)

        print('------------------------------------------------------')
        print(f'|   [StatsPrint]')
        print(f'|   |-- Name: "{c.trainer.name}"')
        print(f'|   |-- Epoch: {c.epoch}')
        print(f'|   |-- Total sample count: {c.sample_count}')
        print(f'|   |-- Total batch count: {c.iteration}')
        print(f'|   |-- Learning rates: {r(current_lrs)}')
        print(f'|   |-- Train')
        print(f'|   |     |-- loss{gdic(t_loss_monitor_result)}')
        print(f'|   |     |     |-- {self._get_print_from_monitor_result(t_loss_monitor_result)}')
        print(f'|   |     |-- metrics')
        print(f'|   |           |-- {gpfm(train_metric_monitor_results, prefix="|   |           |-- ")}')
        print(gpcm(State.TRAIN, c, prefix="\n|   |           |   "))
        print(f'|   |')
        print(f'|   |-- Validation')
        print(f'|         |-- loss{gdic(v_loss_monitor_result)}')
        print(f'|         |     |-- {self._get_print_from_monitor_result(v_loss_monitor_result)}')
        print(f'|         |-- metrics')
        print(f'|               |-- {gpfm(val_metric_monitor_results, prefix="|               |-- ")}')
        print(gpcm(State.VAL, c, prefix="\n|               |   "))
        print('------------------------------------------------------')
        print('') #EMPTY LINE SEPARATOR



import torch as T
import torch.nn as nn
from lpd.extensions.custom_layers import Dense

def examples_data_generator(N, D_in, D_out, binary_out=False, category_out=False):
    # N is batch size
    # D_in is input dimension
    # D_out is output dimension.

    # Create random Tensors to hold inputs and outputs
    x = T.randn(N, D_in)
    if binary_out:
        y = T.randint(low=0,high=2, size=[N, D_out], dtype=T.float)
    elif category_out:
        y = T.randint(low=0,high=D_out, size=[N], dtype=T.long)
    else:
        y = T.randn(N, D_out)
    while True:
        yield x, y #YIELD THE SAME X,y every time

def examples_prediction_data_generator(origin_data_loader, steps):
    for x,y in origin_data_loader:
        steps -= 1
        yield x
        if steps == 0:
            break


def get_basic_model(D_in, H, D_out):
    return nn.Sequential(
                            Dense(D_in, H, use_bias=True, activation=nn.ReLU()),
                            Dense(H, D_out, use_bias=True, activation=None)
                        )


import os
import torch.optim as optim
import torch.nn as nn

from lpd.trainer import Trainer
from lpd.predictor import Predictor
from lpd.callbacks import SchedulerStep, StatsPrint, ModelCheckPoint, LossOptimizerHandler, CallbackMonitor
from lpd.extensions.custom_schedulers import DoNothingToLR
from lpd.enums import Phase, State, MonitorType, StatsType, MonitorMode
from lpd.metrics import BinaryAccuracyWithLogits, MetricBase
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu

gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

save_to_dir = os.path.dirname(__file__) + '/trainer_checkpoint/'
trainer_file_name = 'trainer'

def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 64, 100, 100, 1
    num_epochs = 3
    data_loader = eu.examples_data_generator(N, D_in, D_out, binary_out=True)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps

def get_trainer_base(D_in, H, D_out):
    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.BCEWithLogitsLoss().to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=1e-4)

    scheduler = DoNothingToLR() #CAN ALSO USE scheduler=None, BUT DoNothingToLR IS MORE EXPLICIT
    
    metric_name_to_func = {"Accuracy":BinaryAccuracyWithLogits()}

    return device, model, loss_func, optimizer, scheduler, metric_name_to_func

def get_trainer(N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps):
    device, model, loss_func, optimizer, scheduler, metric_name_to_func = get_trainer_base(D_in, H, D_out)

    callbacks = [   
                    LossOptimizerHandler(),
                    #ADDING ModelCheckPoint WITH save_full_trainer=True TO SAVE FULL TRAINER
                    ModelCheckPoint(checkpoint_dir=save_to_dir, 
                                    checkpoint_file_name=trainer_file_name, 
                                    callback_monitor=CallbackMonitor(patience=None,
                                                                     monitor_type=MonitorType.LOSS, 
                                                                     stats_type=StatsType.VAL, 
                                                                     monitor_mode=MonitorMode.MIN),
                                    save_best_only=True, 
                                    save_full_trainer=True),
                    SchedulerStep(),
                    # SINCE ACCURACY NEEDS TO GO UP AND INACCURACY NEEDS TO GO DOWN, LETS DEFINE CallbackMonitors for StatsPrint PER EACH METRIC
                    StatsPrint(train_metrics_monitors=CallbackMonitor(patience=None, 
                                                                       monitor_type=MonitorType.METRIC,
                                                                       stats_type=StatsType.TRAIN,
                                                                       monitor_mode=MonitorMode.MAX,
                                                                       metric_name='Accuracy'))
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader, 
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Train-Save-Load-Predict-Example')
    return trainer

def run():
    N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps = get_parameters()

    current_trainer = get_trainer(N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps)

    # TRAINING WILL SAVE current_trainer IN ModelCheckPoint
    current_trainer.train(num_epochs)

    # YOU CAN ALSO SAVE THIS TRAINER MANUALLY, LIKE THE CODE BELOW
    current_trainer.save_trainer(save_to_dir, trainer_file_name + '_manual_save')

    # NOW LETS CREATE A PREDICTOR FROM THE SAVED FILE
    device = tu.get_gpu_device_if_available()
    model = eu.get_basic_model(D_in, H, D_out).to(device)
    predictor = Predictor(model=model, device=device)

    data_generator_for_predictions = eu.examples_prediction_data_generator(data_loader, data_loader_steps)

    #PREDICT ON A SINGLE SAMPLE
    sample = next(data_generator_for_predictions)[0]
    sample_prediction = predictor.predict_sample(sample)

    # PREDICT ON A SINGLE BATCH
    batch = next(data_generator_for_predictions)
    batch_prediction = predictor.predict_batch(batch)

    # PREDICTION ON A DATA LOADER
    data_loader_predictions = predictor.predict_data_loader(data_generator_for_predictions, data_loader_steps)


import os
import unittest

import torch.optim as optim
import torch.nn as nn
from lpd.trainer import Trainer
from lpd.callbacks import StatsPrint, SchedulerStep, LossOptimizerHandler, ModelCheckPoint, CallbackMonitor
from lpd.extensions.custom_schedulers import KerasDecay
from lpd.enums import Phase, State, MonitorType, StatsType, MonitorMode
from lpd.metrics import BinaryAccuracyWithLogits, CategoricalAccuracyWithLogits
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu



class TestTrainer(unittest.TestCase):

    def test_metric_name_to_func_validation(self):
        device = tu.get_gpu_device_if_available()

        model = eu.get_basic_model(10, 10, 10).to(device)

        loss_func = nn.BCEWithLogitsLoss().to(device)
    
        optimizer = optim.Adam(model.parameters(), lr=1e-4)

        scheduler = None
        
        metric_name_to_func = {"acc":lambda x,y: x+y}

        callbacks = [   
                        LossOptimizerHandler(),
                        StatsPrint()
                    ]

        data_loader = eu.examples_data_generator(10, 10, 10)
        data_loader_steps = 100

        # ASSERT BAD VALUE FOR metric_name_to_func
        self.assertRaises(ValueError, Trainer, model=model, 
                                                device=device, 
                                                loss_func=loss_func, 
                                                optimizer=optimizer,
                                                scheduler=scheduler,
                                                metric_name_to_func=metric_name_to_func, 
                                                train_data_loader=data_loader, 
                                                val_data_loader=data_loader,
                                                train_steps=data_loader_steps,
                                                val_steps=data_loader_steps,
                                                callbacks=callbacks,
                                                name='Trainer-Test')

        # ASSERT GOOD VALUE FOR metric_name_to_func
        metric_name_to_func = {"acc":BinaryAccuracyWithLogits()}
        trainer = Trainer(model=model, 
                        device=device, 
                        loss_func=loss_func, 
                        optimizer=optimizer,
                        scheduler=scheduler,
                        metric_name_to_func=metric_name_to_func, 
                        train_data_loader=data_loader, 
                        val_data_loader=data_loader,
                        train_steps=data_loader_steps,
                        val_steps=data_loader_steps,
                        callbacks=callbacks,
                        name='Trainer-Test')

    def test_save_and_load(self):
        gu.seed_all(42)
        save_to_dir = os.path.dirname(__file__) + '/trainer_checkpoint/'
        trainer_file_name = 'trainer'

        device = tu.get_gpu_device_if_available()

        model = eu.get_basic_model(10, 10, 10).to(device)

        loss_func = nn.CrossEntropyLoss().to(device)
    
        optimizer = optim.Adam(model.parameters(), lr=1e-4)

        scheduler = KerasDecay(optimizer, 0.0001, last_step=-1)
        
        metric_name_to_func = {"acc":CategoricalAccuracyWithLogits()}

        callbacks = [   
                        LossOptimizerHandler(),
                        ModelCheckPoint(checkpoint_dir=save_to_dir, 
                                        checkpoint_file_name=trainer_file_name, 
                                        callback_monitor=CallbackMonitor(patience=-1,
                                                                         monitor_type=MonitorType.LOSS, 
                                                                         stats_type=StatsType.VAL, 
                                                                         monitor_mode=MonitorMode.MIN),
                                        save_best_only=False, 
                                        save_full_trainer=True,
                                        verbose=0),
                        SchedulerStep(apply_on_phase=Phase.BATCH_END, apply_on_states=State.TRAIN),
                        StatsPrint()
                    ]

        
        data_loader = eu.examples_data_generator(10, 10, 10, category_out=True)
        data_loader_steps = 100
        num_epochs = 5

        trainer = Trainer(model=model, 
                        device=device, 
                        loss_func=loss_func, 
                        optimizer=optimizer,
                        scheduler=scheduler,
                        metric_name_to_func=metric_name_to_func, 
                        train_data_loader=data_loader, 
                        val_data_loader=data_loader,
                        train_steps=data_loader_steps,
                        val_steps=data_loader_steps,
                        callbacks=callbacks,
                        name='Trainer-Test')
        
        trainer.train(num_epochs, verbose=0)

        loaded_trainer = Trainer.load_trainer(dir_path=save_to_dir,
                                            file_name=trainer_file_name + f'_epoch_{num_epochs}',
                                            model=model,
                                            device=device,
                                            loss_func=loss_func,
                                            optimizer=optimizer,
                                            scheduler=scheduler,
                                            train_data_loader=data_loader, 
                                            val_data_loader=data_loader,
                                            train_steps=data_loader_steps,
                                            val_steps=data_loader_steps)
        
        self.assertEqual(loaded_trainer.epoch, trainer.epoch)
        self.assertListEqual(tu.get_lrs_from_optimizer(loaded_trainer.optimizer), tu.get_lrs_from_optimizer(trainer.optimizer))
        self.assertEqual(loaded_trainer.callbacks[1].monitor._get_best(), trainer.callbacks[1].monitor._get_best())

    def test_loss_handler_validation(self):
        device = tu.get_gpu_device_if_available()

        model = eu.get_basic_model(10, 10, 10).to(device)

        loss_func = nn.BCEWithLogitsLoss().to(device)
    
        optimizer = optim.Adam(model.parameters(), lr=1e-4)

        scheduler = KerasDecay(optimizer, 0.0001, last_step=-1)
        
        metric_name_to_func = {"acc":BinaryAccuracyWithLogits()}

        callbacks = [   
                        StatsPrint()
                    ]

        
        data_loader = eu.examples_data_generator(10, 10, 10)
        data_loader_steps = 100
        num_epochs = 5
        verbose = 0
        
        trainer = Trainer(model=model, 
                        device=device, 
                        loss_func=loss_func, 
                        optimizer=optimizer,
                        scheduler=scheduler,
                        metric_name_to_func=metric_name_to_func, 
                        train_data_loader=data_loader, 
                        val_data_loader=data_loader,
                        train_steps=data_loader_steps,
                        val_steps=data_loader_steps,
                        callbacks=callbacks,
                        name='Trainer-Test')
        
        self.assertRaises(ValueError, trainer.train, num_epochs, verbose)

        trainer.callbacks.append(LossOptimizerHandler())
        trainer.train(num_epochs, verbose=0)



import torch as T
from lpd.enums.confusion_matrix_based_metric import ConfusionMatrixBasedMetric as metric

class ConfusionMatrix():
    """
        Agrs:
            num_classes - The number of classes in the classification
            labels - names of classes, for nice prints, if not provided, the class index will be the label
            predictions_to_classes_convertor - (optional) a function that takes y_pred batch and y_true batch and converts it into class indices batch where
                                               each index represents the chosen class
                                               if None:  
                                                        torch.max with indices will be used for multi-class .
                                                        threshold will be used for binary or multilabel.
            threshold - for binary or multilable classification

    """
    def __init__(self, num_classes, labels=None, predictions_to_classes_convertor=None, threshold=0.5):
        self.num_classes = num_classes
        self.class_idxs = list(range(self.num_classes))
        self.labels = self._build_labels(labels)
        self.class_idx_to_label = self._build_idx_to_label()
        self.confusion = self._build_confusion()
        self.threshold = threshold
        self.predictions_to_classes_convertor = predictions_to_classes_convertor if predictions_to_classes_convertor else self._convert_predictions_to_classes
        if self.predictions_to_classes_convertor is None:
            raise ValueError(f'[ConfusionMatrix] - Got None for predictions_to_classes_convertor, expected a function')

        self.last_stats = None

    def _build_labels(self, labels):
        if labels is None:
            return [str(c) for c in self.class_idxs]
        return labels

    def _build_idx_to_label(self):
        return dict(zip(self.class_idxs, self.labels)) 
    
    def _build_confusion(self):
        length = len(self.class_idxs)
        confusion = T.zeros((length, length)).long()
        return confusion

    def _normalized(self):
        asum = T.sum(self.confusion)
        return self.confusion / asum.float()

    def _convert_predictions_to_classes(self, y_pred: T.Tensor, y_true: T.Tensor):
        if not (len(y_pred.shape) == len(y_true.shape) or \
                len(y_pred.shape) == len(y_true.shape) + 1):
            raise ValueError("Expecting y_pred and y_true to has same amount of dimentions, or y_pred to have an extra one")

        if self.num_classes == 2:
            if len(y_pred.shape) == 2 and y_pred.shape[1] == 1:
                # BINARY AS SINGLE DIGIT WRAPPED
                classes = (y_pred.squeeze(1) >= self.threshold).long()
                return classes
            if len(y_pred.shape) == 1:
                # BINARY AS SINGLE DIGIT
                classes = (y_pred >= self.threshold).long()
                return classes

        if len(y_pred.shape) == len(y_true.shape) + 1:
            # MULTI CLASS
            classes = T.max(y_pred, dim=1)[1]
            return classes

    def get_confusion_matrix(self, normalized=False):
        if normalized:
            return self._normalized().clone().detach()
        return self.confusion.clone().detach()

    def get_stats(self):
        if self.last_stats:
            # USE CACHE
            return self.last_stats

        asum = T.sum(self.confusion)
        stats = {}
        for class_idx in self.class_idxs:
            class_stats = {}
            row_sum = T.sum(self.confusion[class_idx])
            col_sum = T.sum(self.confusion[:,class_idx])

            tp  = self.confusion[class_idx][class_idx]
            fp  = row_sum - tp
            fn  = col_sum - tp
            tn  = asum - row_sum - col_sum + tp  # +TP BECAUSE ROW_SUM AND COL_SUM REMOVES self.confusion[class_idx][class_idx] TWICE

            class_stats[metric.TP] = tp
            class_stats[metric.FP] = fp
            class_stats[metric.FN] = fn
            class_stats[metric.TN] = tn


            # SUMS
            tp_fp = (tp + fp).float()
            tp_fn = (tp + fn).float()
            tp_tn = (tp + tn).float()
            tn_fn = (tn + fn).float()
            fp_fn = (fp + fn).float()
            tn_fp = (tn + fp).float()

            class_stats[metric.PRECISION]                    = (tp / tp_fp) if tp_fp > 0 else 0.0
            class_stats[metric.SENSITIVITY]                  = (tp / tp_fn) if tp_fn > 0 else 0.0
            class_stats[metric.SPECIFICITY]                  = (tn / tn_fp) if tn_fp > 0 else 0.0
            class_stats[metric.RECALL]                       = (tp / tp_fn) if tp_fn > 0 else 0.0
            class_stats[metric.POSITIVE_PREDICTIVE_VALUE]    = (tp / tp_fp) if tp_fp > 0 else 0.0
            class_stats[metric.NEGATIVE_PREDICTIVE_VALUE]    = (tn / tn_fn) if tn_fn > 0 else 0.0
            class_stats[metric.ACCURACY]                     = (tp_tn / asum) if asum > 0 else 0.0
            class_stats[metric.F1SCORE]                      = ((2 * tp) / ((2 * tp) + fp_fn)) if ((2 * tp) + fp_fn) > 0 else 0.0


            stats[self.class_idx_to_label[class_idx]] = class_stats
        self.last_stats = stats
        return stats

    def confusion_matrix_string(self, row_prefix='\n', normalized=False):
        max_value_len = len(str(T.max(self.confusion).item()))
        cellwidth = max(7, max_value_len)

        if normalized:
            confusion = self._normalized()
        else:
            confusion = self.confusion

        def handle_confusion_row(confusion_row):
            if normalized:
                return [("%0.2f%%" % (val * 100)).ljust(cellwidth)[0:cellwidth] for val in confusion_row]
            else:
                return [str(val.item()).ljust(cellwidth)[0:cellwidth] for val in confusion_row]

        cm_strings = [f'Confusion Matrix, ᴾ = Pred, ᵀ = True', '']

        cm_strings.append("        %s" % " ".join([("%sᵀ" % label).ljust(cellwidth)[0:cellwidth] for label in self.labels]))
        for label, confusion_row in zip(self.labels, confusion):
            cm_strings.append("%s %s" % (("%sᴾ" % label).ljust(cellwidth)[0:cellwidth], " ".join(handle_confusion_row(confusion_row))))

        cm_strings.append('')
        return row_prefix.join(cm_strings)

    def reset(self):
        self.last_stats = None # INVALIDATE CACHE
        self.confusion = self.confusion * 0

    def update_state(self, y_pred: T.Tensor, y_true: T.Tensor):
        self.last_stats = None # INVALIDATE CACHE
        y_pred_class_idxs = self.predictions_to_classes_convertor(y_pred, y_true)
        y_true_class_idxs = y_true.long()

        for row, col in zip(y_pred_class_idxs, y_true_class_idxs):
            self.confusion[row][col] += 1



from lpd.enums import Phase, State
from lpd.callbacks.callback_base import CallbackBase
from typing import Union, List, Optional


class LossOptimizerHandlerBase(CallbackBase):
    """
        In case LossOptimizerHandler does not suitable for your needs, create your custom
        callback and derive from this class, and implement __call__ .
        There you have full control for handling loss and optimizer

        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            round_values_on_print_to - see in CallbackBase
    """


    def __init__(self, apply_on_phase: Phase, 
                    apply_on_states: Union[State, List[State]], 
                    round_values_on_print_to: Optional[int]=None):
        super(LossOptimizerHandlerBase, self).__init__(apply_on_phase, apply_on_states, round_values_on_print_to)


class CallbackContext():
    #REPRESENTS THE INPUT TO THE CALLBACK, NOTICE, SOME VALUES MIGHT BE NONE, DEPENDING ON THE PHASE/STATE OF THE CALLBACK
    def __init__(self, trainer):
        self.epoch = trainer.epoch
        self.sample_count = trainer.sample_count
        self.sample_count_in_epoch = trainer.sample_count_in_epoch
        self.iteration = trainer.iteration
        self.iteration_in_epoch = trainer.iteration_in_epoch

        self.train_stats = trainer.train_stats
        self.val_stats = trainer.val_stats
        self.test_stats = trainer.test_stats

        self.trainer_state = trainer.state
        self.trainer_phase = trainer.phase
        self.scheduler = trainer.scheduler
        self.optimizer = trainer.optimizer
        self.trainer = trainer

        self.train_last_loss = trainer.train_last_loss



import os
import torch.optim as optim
import torch.nn as nn

from lpd.trainer import Trainer
from lpd.callbacks import SchedulerStep, StatsPrint, ModelCheckPoint, LossOptimizerHandler, CallbackMonitor
from lpd.extensions.custom_schedulers import DoNothingToLR
from lpd.enums import Phase, State, MonitorType, StatsType, MonitorMode, MetricMethod
from lpd.metrics import BinaryAccuracyWithLogits, MetricBase
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu

gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

save_to_dir = os.path.dirname(__file__) + '/trainer_checkpoint/'
trainer_file_name = 'trainer'


def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 64, 100, 100, 1
    num_epochs = 5
    data_loader = eu.examples_data_generator(N, D_in, D_out, binary_out=True)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps


# LET'S CREATE A CUSTOM (ALTOUGH NOT SO INFORMATIVE) METRIC
class InaccuracyWithLogits(MetricBase):
    def __init__(self):
        super(InaccuracyWithLogits, self).__init__(MetricMethod.MEAN)
        self.bawl = BinaryAccuracyWithLogits() # we exploit BinaryAccuracyWithLogits for the computation

    def __call__(self, y_pred, y_true): # <=== implement this method!
        acc = self.bawl(y_pred, y_true)
        return 1 - acc  # return the inaccuracy


def get_trainer_base(D_in, H, D_out):
    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.BCEWithLogitsLoss().to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=1e-4)

    scheduler = DoNothingToLR() #CAN ALSO USE scheduler=None, BUT DoNothingToLR IS MORE EXPLICIT
    
    metric_name_to_func = {"Accuracy":BinaryAccuracyWithLogits(), "InAccuracy":InaccuracyWithLogits()}

    return device, model, loss_func, optimizer, scheduler, metric_name_to_func

def get_trainer(N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps):
    device, model, loss_func, optimizer, scheduler, metric_name_to_func = get_trainer_base(D_in, H, D_out)

    callbacks = [   
                    LossOptimizerHandler(),
                    #ADDING ModelCheckPoint WITH save_full_trainer=True TO SAVE FULL TRAINER
                    ModelCheckPoint(checkpoint_dir=save_to_dir, 
                                    checkpoint_file_name=trainer_file_name, 
                                    callback_monitor=CallbackMonitor(patience=None,
                                                                     monitor_type=MonitorType.LOSS, 
                                                                     stats_type=StatsType.VAL, 
                                                                     monitor_mode=MonitorMode.MIN),
                                    save_best_only=True, 
                                    save_full_trainer=True),
                    SchedulerStep(),
                    # SINCE ACCURACY NEEDS TO GO UP AND INACCURACY NEEDS TO GO DOWN, LETS DEFINE CallbackMonitors for StatsPrint PER EACH METRIC
                    StatsPrint(train_metrics_monitors=[CallbackMonitor(patience=None, 
                                                                       monitor_type=MonitorType.METRIC,
                                                                       stats_type=StatsType.TRAIN,
                                                                       monitor_mode=MonitorMode.MAX,
                                                                       metric_name='Accuracy'),
                                                       CallbackMonitor(patience=None, 
                                                                       monitor_type=MonitorType.METRIC,
                                                                       stats_type=StatsType.TRAIN,
                                                                       monitor_mode=MonitorMode.MIN,
                                                                       metric_name='InAccuracy')])
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader, 
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Save-And-Load-Example')
    return trainer

def load_trainer(N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps):
    device, model, loss_func, optimizer, scheduler, metric_name_to_func = get_trainer_base(D_in, H, D_out)

    # NOTICE, load_trainer IS A STATIC METHOD IN Trainer CLASS
    loaded_trainer = Trainer.load_trainer(dir_path=save_to_dir,
                                            file_name=trainer_file_name + '_manual_save',
                                            model=model,
                                            device=device,
                                            loss_func=loss_func,
                                            optimizer=optimizer,
                                            scheduler=scheduler,
                                            train_data_loader=data_loader, 
                                            val_data_loader=data_loader,
                                            train_steps=data_loader_steps,
                                            val_steps=data_loader_steps)
    return loaded_trainer

def run():
    params = get_parameters()
    num_epochs = params[4]
    current_trainer = get_trainer(*params)

    current_trainer.summary()

    # TRAINING WILL SAVE current_trainer IN ModelCheckPoint
    current_trainer.train(num_epochs)

    # YOU CAN ALSO SAVE THIS TRAINER MANUALLY, LIKE THE CODE BELOW
    current_trainer.save_trainer(save_to_dir, trainer_file_name + '_manual_save')

    # NOW LETS CREATE A NEW TRAINER FROM THE SAVED FILE,
    loaded_trainer = load_trainer(*params)

    # CONTINUE TRAINING
    loaded_trainer.train(num_epochs)



from lpd.metrics.metric_base import MetricBase, MetricConfusionMatrixBase
from lpd.metrics.binary_accuracy import BinaryAccuracy
from lpd.metrics.binary_accuracy_with_logits import BinaryAccuracyWithLogits
from lpd.metrics.categorical_accuracy import CategoricalAccuracy
from lpd.metrics.categorical_accuracy_with_logits import CategoricalAccuracyWithLogits
from lpd.metrics.top_k_categorical_accuracy import TopKCategoricalAccuracy
from lpd.metrics.tp_tn_fp_fn import TruePositives, FalsePositives, TrueNegatives, FalseNegatives




from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitor, CallbackMonitorResult
from typing import Union, List, Optional, Dict

class CollectOutputs(CallbackBase):
    """
        This callback will collect outputs per each state, (it is currently used in trainer.predict() method.)
        It will collect the numpy outputs in the defined states to a dictionary (state->outputs)

        Methods:
            get_outputs_for_state - for a given state, returns the collected outputs

        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
    """

    def __init__(self, 
                    apply_on_phase: Phase, 
                    apply_on_states: Union[State, List[State]]):
        super(CollectOutputs, self).__init__(apply_on_phase, apply_on_states)
        self.state_to_outputs = {}

    def get_outputs_for_state(self, state: State):
        return [data.cpu().numpy() for data in self.state_to_outputs[state]]

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD
        state = c.trainer_state

        if self.should_apply_on_state(c):

            if state not in self.state_to_outputs:
                self.state_to_outputs[state] = []

            last_outputs = c.trainer._last_outputs.data
            self.state_to_outputs[state].append(last_outputs)



from enum import Enum, auto

class MetricMethod(Enum):
    """
        use this to dictate how a metric behaves over the batches,
        whether its accumulates the MEAN, or the SUM, or taking the LAST value (for example in MetricConfusionMatrixBase)
    """
    MEAN    = auto()
    SUM     = auto() 
    LAST    = auto() 
    
    def __str__(self):
        return self.name


# IN THIS EXAMPLE WE WILL USE THE PYTORCH DATALOADER (AS OPPOSE TO PYTHON GENERATORS)

import torch as T
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader

from lpd.trainer import Trainer
from lpd.extensions.custom_layers import Dense
from lpd.metrics import BinaryAccuracyWithLogits
from lpd.callbacks import StatsPrint, EarlyStopping, SchedulerStep, LossOptimizerHandler, CallbackMonitor
from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
import lpd.utils.general_utils as gu
import lpd.utils.torch_utils as tu


num_embeddings = 10
num_samples_per_file = 100
def generate_file_data():
    # [(X,y), (X,y), (X,y), (X,y)] like [ (1,0), (5,1), (6,1) ...]
    return [(T.randint(1,num_embeddings,(1,1)).squeeze(), T.randint(0,2,(1,1)).squeeze().float()) for _ in range(num_samples_per_file)]


class MyDataset(Dataset):
    def __init__(self, file_ids):
        self.file_ids = file_ids
        self.idx_to_file_id = {idx:file_id for idx, file_id in enumerate(self.file_ids)}
        self.file_id_to_sample_idx = {i:0 for i in self.file_ids}

    def __len__(self):
        return len(self.file_ids)

    def __getitem__(self, index):
        file_id = self.idx_to_file_id[index]
        
        data = raw_data[file_id]
        sample_idx = self.file_id_to_sample_idx[file_id]
        self.file_id_to_sample_idx[file_id] = (self.file_id_to_sample_idx[file_id] + 1) % num_samples_per_file

        data_at_sample_idx = data[sample_idx]

        X = data_at_sample_idx[0]
        y = data_at_sample_idx[1]

        return X, y

# Parameters
params = {  'data_loader_params': { 
                                    'batch_size': 8,
                                    'shuffle': True,
                                    'num_workers': 1
                                   },
            'H': 128,
            'D_out': 1,
            'embedding_dim': 64,
            'num_epochs': 80}

# Datasets
file_ids = [f'id-{i}' for i in range(16)]
partition = {
                #FOR SIMPLICITY, TRAIN/VAL/TEST WILL BE THE SAME DATA
                'train': file_ids, 
                'val':   file_ids,  
                'test':  file_ids   
            }

raw_data = {file_id:generate_file_data() for file_id in partition['train']}

# Generators
train_dataset = MyDataset(partition['train'])
train_data_loader = DataLoader(train_dataset, **params['data_loader_params'])

val_dataset = MyDataset(partition['val'])
val_data_loader = DataLoader(val_dataset, **params['data_loader_params'])

test_dataset = MyDataset(partition['test'])
test_data_loader = DataLoader(test_dataset, **params['data_loader_params'])


class Model(nn.Module):
    def __init__(self, H, D_out, num_embeddings, embedding_dim):
        super(Model, self).__init__()

        #LAYERS

        self.embedding_layer = nn.Embedding(num_embeddings=num_embeddings + 1, 
                                            embedding_dim=embedding_dim)

        self.dense = Dense(embedding_dim, H, use_bias=True, activation=nn.ReLU())
        self.dense_out = Dense(H, D_out, use_bias=True, activation=None)

    def forward(self, x):               
        x = self.embedding_layer(x)     # (batch, embedding_dim)
        x = self.dense(x)               # (batch, H)
        x = self.dense_out(x)           # (batch, 1)
        x = x.squeeze()                 # (batch)
        return x #NOTICE! LOGITS OUT, NOT SIGMOID, THE SIGMOID WILL BE APPLIED IN BCEWithLogitsLoss

def get_trainer(params):

    device = tu.get_gpu_device_if_available()

    # Use the nn package to define our model and loss function.
    model = Model(params['H'], params['D_out'], num_embeddings, params['embedding_dim']).to(device)

    loss_func = nn.BCEWithLogitsLoss().to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=0.1)

    scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.99)

    metric_name_to_func = {"acc":BinaryAccuracyWithLogits()}

    callbacks = [  
                    LossOptimizerHandler(),
                    SchedulerStep(apply_on_phase=Phase.BATCH_END, apply_on_states=State.TRAIN),
                    EarlyStopping(callback_monitor=CallbackMonitor(patience=3, monitor_type=MonitorType.LOSS, stats_type=StatsType.VAL, monitor_mode=MonitorMode.MIN)),
                    StatsPrint(round_values_on_print_to=7)
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=train_data_loader, 
                      val_data_loader=val_data_loader,
                      train_steps=len(train_dataset),
                      val_steps=len(val_dataset),
                      callbacks=callbacks,
                      name='DataLoader-Example')
    return trainer


def run():
    gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

    trainer = get_trainer(params)
    
    trainer.summary()

    trainer.train(params['num_epochs'])

    evaluation = trainer.evaluate(test_data_loader, len(test_data_loader))

    print(evaluation)


import torch as T
from lpd.metrics.metric_base import MetricBase
from lpd.metrics.categorical_accuracy import CategoricalAccuracy
from lpd.enums.metric_method import MetricMethod


class CategoricalAccuracyWithLogits(MetricBase):
    """
        Same as CategoricalAccuracy, but more explicit about the Logits
    """
    def __init__(self):
        super(CategoricalAccuracyWithLogits, self).__init__(MetricMethod.MEAN)
        self.ca = CategoricalAccuracy()

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        return self.ca(y_pred, y_true)


import os
import torch.optim as optim
import torch.nn as nn

from lpd.trainer import Trainer
from lpd.callbacks import SchedulerStep, StatsPrint, ModelCheckPoint, LossOptimizerHandler, CallbackMonitor
from lpd.extensions.custom_schedulers import DoNothingToLR
from lpd.enums import Phase, State, MonitorType, StatsType, MonitorMode
from lpd.metrics import TruePositives, FalsePositives, TrueNegatives, FalseNegatives
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu

gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out, num_classes = 128, 100, 100, 3,3
    num_epochs = 5
    data_loader = eu.examples_data_generator(N, D_in, D_out, category_out=True)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, num_classes, data_loader, data_loader_steps


def get_trainer_base(D_in, H, D_out, num_classes):
    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.CrossEntropyLoss().to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=1e-4)

    scheduler = DoNothingToLR() #CAN ALSO USE scheduler=None, BUT DoNothingToLR IS MORE EXPLICIT
    
    labels = ['Cat', 'Dog', 'Bird']
    metric_name_to_func = {
                            "TP":TruePositives(num_classes, labels=labels, threshold = 0),
                            "FP":FalsePositives(num_classes, labels=labels, threshold = 0),
                            "TN":TrueNegatives(num_classes, labels=labels, threshold = 0),
                            "FN":FalseNegatives(num_classes, labels=labels, threshold = 0)
                          }

    return device, model, loss_func, optimizer, scheduler, metric_name_to_func


def get_trainer(N, D_in, H, D_out, num_epochs, num_classes, data_loader, data_loader_steps):
    device, model, loss_func, optimizer, scheduler, metric_name_to_func = get_trainer_base(D_in, H, D_out, num_classes)

    callbacks = [   
                    LossOptimizerHandler(),
                    StatsPrint(print_confusion_matrix=True)
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader, 
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Confusion-Matrix-Example')
    return trainer


def run():
    N, D_in, H, D_out, num_epochs, num_classes, data_loader, data_loader_steps = get_parameters()

    current_trainer = get_trainer(N, D_in, H, D_out, num_epochs, num_classes, data_loader, data_loader_steps)

    current_trainer.train(num_epochs)




class Config():
	def __init__(self):
		self.TENSORBOARD_DIR = './tensorboard/'

		self.LEARNING_RATE = 0.1
		self.STEP_LR_STEP_SIZE = 1
		self.STEP_LR_GAMMA = 0.9

		self.BATCH_SIZE = 512
		self.NUM_EPOCHS = 100
		self.EARLY_STOPPING_PATIENCE = 2

		self.EMBEDDINGS_SIZE = 8										

		#TRANSFORMER
		self.NUM_TRANSFORMER_ENCODERS = 1								
		self.NUM_HEADS_PER_TRANSFORMER = 2 								
		self.TRANSFORMER_DROP_OUT_PROBA = 0.1							
		self.TRANSFORMER_KEY_DIM = self.EMBEDDINGS_SIZE // 2            
		self.TRANSFORMER_FF_EXPANSION_RATE = 1							

		self.MODEL_WEIGHTS_FILE_NAME = 'weights'
		self.MODEL_WEIGHTS_DIR = './weights/'

		#IN DATA GENERATOR TUPLE
		self.IDX_OF_X1 = 0
		self.IDX_OF_X2 = 1
		self.IDX_OF_X3 = 2
		self.IDX_OF_LABEL = 3 




from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitor, CallbackMonitorResult
import lpd.utils.file_utils as fu
from lpd.utils.torch_utils import save_checkpoint
from typing import Union, List, Optional, Dict

class ModelCheckPoint(CallbackBase):
    """
        Saving a checkpoint when a monitored loss/metric has improved.
        Checkpoint will save the model, optimizer, scheduler and epoch number.
        You can also configure it to save Full Trainer.

        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            checkpoint_dir - the folder to dave the model, if None was passed, will use current folder
            checkpoint_file_name - the name of the file that will be saved
            callback_monitor - to monitor the loss/metric for saving checkpoint when improved
            save_best_only - if True, will override previous best model, else, will keep both
            verbose - 0=no print, 1=print
            round_values_on_print_to - see in CallbackBase
            save_full_trainer - if True, will save all trainer parameters to be able to continue where you left off
    """

    def __init__(self,  apply_on_phase: Phase=Phase.EPOCH_END, 
                        apply_on_states: Union[State, List[State]]=State.EXTERNAL,
                        checkpoint_dir: str=None, 
                        checkpoint_file_name: str='checkpoint', 
                        callback_monitor: CallbackMonitor=None,
                        save_best_only: bool=False, 
                        verbose: int=1,
                        round_values_on_print_to: int=None,
                        save_full_trainer: bool=False):
        super(ModelCheckPoint, self).__init__(apply_on_phase, apply_on_states, round_values_on_print_to)
        if checkpoint_dir is None:
            raise ValueError("[ModelCheckPoint] - checkpoint_dir was not provided")
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_file_name = checkpoint_file_name
        if callback_monitor is None:
            raise ValueError("[ModelCheckPoint] - callback_monitor was not provided")
        self.monitor = callback_monitor
        self.save_best_only = save_best_only
        self.verbose = verbose  # VERBOSITY MODE, 0 OR 1.
        self.save_full_trainer = save_full_trainer
        self._ensure_folder_created()

    def _ensure_folder_created(self):
        if not fu.is_folder_exists(self.checkpoint_dir):
            fu.create_folder(self.checkpoint_dir)

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD
        r = self.round_to #READABILITY DOWN THE ROAD

        monitor_result = self.monitor.track(callback_context)
        if monitor_result.has_improved():
            msg = f'[ModelCheckPoint] - {monitor_result.description} improved from {r(monitor_result.prev_best)} to {r(monitor_result.new_best)}'
            if self.save_best_only:
                file_name = f'{self.checkpoint_file_name}_best_only'
            else:
                file_name = f'{self.checkpoint_file_name}_epoch_{c.epoch}'
            
            if self.save_full_trainer:
                c.trainer.save_trainer(self.checkpoint_dir, file_name, msg=msg, verbose=self.verbose)
            else:
                save_checkpoint(self.checkpoint_dir, file_name, c.trainer, msg=msg, verbose=self.verbose)
        else:
            if self.verbose:
                print(f'[ModelCheckPoint] - {monitor_result.description} did not improved from {monitor_result.prev_best}.')



import torch.optim as optim
import torch.nn as nn

from lpd.trainer import Trainer
from lpd.callbacks import StatsPrint, SchedulerStep, LossOptimizerHandler
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu


def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 1, 1000, 100, 10
    num_epochs = 10
    data_loader = eu.examples_data_generator(N, D_in, D_out)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps

def get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps):

    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.MSELoss(reduction='sum').to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=1e-4)

    metric_name_to_func = None # THIS EXAMPLE DOES NOT USE METRICS, ONLY LOSS

    # HERE WE DEFINE A CLOSURE THAT WILL ACCUMULATE BATCHES UNTIL WE REACH AT LEAST 32 SAMPLES
    # WITH GRADIENTS, AND ONLY THEN, INVOKE STEP AND ZERO GRAD
    # SINCE IN THIS EXAMPLE THE BATCH SIZE WAS SET TO 1, WE EXPECT 31 SKIPS AND INVOCATION ON THE BATCH AFTER
    # NOTICE, WE WILL ALSO INVOKE STEP AND ZERO GRAD ON THE LAST BATCH 
    def get_optimizer_handler(action: str):
        invoke_with_min_samples = 32
        sample_count = 0

        def is_last_batch(trainer):
            return trainer.iteration_in_epoch == trainer.train_steps

        def handler(callback_context):
            trainer = callback_context.trainer
            opt = callback_context.optimizer
            nonlocal sample_count
            invoke = False
            if trainer.sample_count_in_epoch - sample_count >= invoke_with_min_samples:
                sample_count = trainer.sample_count_in_epoch
                invoke = True
                print(f'[Trainer] - Accumulated {invoke_with_min_samples} samples, invoking optimizer {action}')

            if is_last_batch(trainer):
                sample_count = 0 #PREPARE FOR NEXT EPOCH
                invoke = True
                print(f'[Trainer] - Last batch in epoch, invoking optimizer {action}')
            
            if invoke:
                if action == 'step':
                    opt.step()
                elif action == 'zero_grad':
                    opt.zero_grad()
            else:
                print(f'[Trainer] - SKIPPING optimizer {action}, {(invoke_with_min_samples - (trainer.sample_count_in_epoch - sample_count))} samples to go')

        return handler

    callbacks = [  
                    LossOptimizerHandler(optimizer_step_handler=get_optimizer_handler('step'), optimizer_zero_grad_handler=get_optimizer_handler('zero_grad')), 
                    StatsPrint()
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=None,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader,  # DATA LOADER WILL YIELD BATCH SIZE OF 1
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Accumulate-Grads-Example')
    return trainer

def run():
    gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

    N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps = get_parameters()

    trainer = get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps)

    trainer.train(num_epochs)


import random
import torch as T
import os 
from .config import Config
from .model import get_trainer
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu


def prepare_chunk_to_model_input(config, chunk):
    x1 = [c[config.IDX_OF_X1] for c in chunk]
    x2 = [c[config.IDX_OF_X2] for c in chunk]
    x3 = [c[config.IDX_OF_X3] for c in chunk]
    y = [c[config.IDX_OF_LABEL] for c in chunk]
    return [T.LongTensor(x1), T.LongTensor(x2), T.LongTensor(x3)], T.Tensor(y)

def get_data_stats(data_generator, verbose=1):
    sanity_count = int(1e6)

    steps = 0
    num_positive_sampling = 0
    num_negative_sampling = 0
    for _, y in data_generator:
        steps += 1
        for y_value in y:
            if y_value == 0:
                num_negative_sampling += 1
            else:
                num_positive_sampling += 1
        if steps > sanity_count:
            # IF ITS AN INF GENERATOR, WE NEED TO STOP EVENTUALLY
            break

    return {'num_positive_sampling': num_positive_sampling,
            'num_negative_sampling': num_negative_sampling,
            'steps': steps,
            'did_generator_finished': (steps < sanity_count)}

def data_generator(config, num_embeddings, num_cycles=1e9):
    # data_generator WILL CREATE FAKE DATA
    # EACH SAMPLE CONTAINS 3 INPUTS: X1,X2,X3, WHERE 
    # X1 IS A SEQUENCE OF EMBEDDING INDICES (LIKE A SENTENCE)
    # X2 IS SINGLE EMBEDDING INDEX          (LIKE A SINGLE WORD)
    # X3 IS SINGLE EMBEDDING INDEX          (LIKE A SINGLE WORD)
    # y IS 0/1 - SO ITS LIKE A BINARY CLASSIFICATION CHALLENGE

    emb_idxs = list(range(num_embeddings))

    seq_length_to_samples = {}
    for x1_seq_len in range(4, 20):  # X1, SEQ LEN WILL RANGE FROM 4-20, SO NOT ALL SEQUENCES ARE OF THE SAME LENGHT

        # ALL THE SEQUENCES OF LEN x1_seq_len WILL SHARE THE SAME ENTRY IN THE seq_length_to_samples DICTIONARY
        seq_length_to_samples[x1_seq_len] = []
        for s in range(1, 100):  # 100 SAMPLES PER EACH LEN
            choice = random.sample(emb_idxs, x1_seq_len + 2)
            x1 = choice[:-2]    # TAKE ALL EMBEDDING INDICES BUT THE LAST 2
            x2 = choice[-2]     # TAKE THE SECOND TO LAST EMBEDDING INDEX
            x3 = choice[-1]     # TAKE THE LAST EMBEDDING INDEX
            y = s % 2           # TOGGLE 0/1/0/1/0/1...
            seq_length_to_samples[x1_seq_len].append((x1, x2, x3, y)) # APPEND FULL SINGLE SAMPLE

    # NOW THAT WE HAVE DATA, WE CAN YIELD IT WITH A SIMPLE GENERATOR
    while num_cycles > 0:
        num_cycles -= 1
        for seq_length, samples in seq_length_to_samples.items():
            yield prepare_chunk_to_model_input(config, samples)


# -----------------[RUN]----------------- #


def run():
    base_path = os.path.dirname(__file__) + '/'
    config = Config()
    tensorboard_data_dir = base_path + config.TENSORBOARD_DIR
    model_weights_dir = base_path + config.MODEL_WEIGHTS_DIR

    num_embeddings = 10000  # ITS FAKE DATA, SO...
    gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

    data_stats = get_data_stats(data_generator(config, num_embeddings, num_cycles=1))
    print(data_stats)
    steps = data_stats['steps']

    generator_infinite = data_generator(config, num_embeddings)

    trainer = get_trainer(config,
                          num_embeddings,
                          generator_infinite,
                          generator_infinite,
                          steps,
                          steps,
                          model_weights_dir,
                          config.MODEL_WEIGHTS_FILE_NAME,
                          tensorboard_data_dir)
    trainer.summary()

    trainer.train(config.NUM_EPOCHS)

    evaluation = trainer.evaluate(generator_infinite, steps)
    print(evaluation)



import torch as T
from lpd.metrics.metric_base import MetricBase
from lpd.enums.metric_method import MetricMethod


class BinaryAccuracy(MetricBase):
    def __init__(self, threshold=0.5):
        super(BinaryAccuracy, self).__init__(MetricMethod.MEAN)
        self.threshold = threshold

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        assert y_true.size() == y_pred.size()
        pred = y_pred >= self.threshold
        accuracy = pred.long().eq(y_true).float().sum() / y_true.numel()
        return accuracy





import os
import pickle
import gzip
from shutil import copytree, copyfile, move

def is_folder_exists(folder_path):
    return os.path.exists(folder_path)

def create_folder(full_path):
    os.makedirs(full_path)

def delete_file(full_path):
    os.remove(full_path)

def copy_paste_file(source_path, dest_path):
    copyfile(source_path, dest_path)

def copy_paste_folder(source_path, dest_path, only_content = False):
    if only_content:
        for root, dirs, files in os.walk(source_path):
            for file_name in files:
                copyfile(source_path + file_name, dest_path + file_name)
            for dir_name in dirs:
                copytree(source_path + dir_name, dest_path)
    else:
        #WHOLE DIRECTORY
        copytree(source_path, dest_path)

def move_folder_location(source_path, dest_path):
    move(source_path, dest_path)

def move_file_location(source_path, dest_path):
    move(source_path, dest_path)

def load_object_from_file(file_path, verbose = 1):
    with_pickle_suffix = file_path + '.pickle'
    if verbose == 1:
        print('loading object from file:', with_pickle_suffix)
    with open(with_pickle_suffix, 'rb') as f:
        obj = pickle.load(f)
    return obj

def save_object_as_file(obj, file_path):
    with_pickle_suffix = file_path + '.pickle'
    print('saving object to file:', with_pickle_suffix)
    with open(with_pickle_suffix, 'wb') as file:
        pickle.dump(obj, file, protocol=pickle.HIGHEST_PROTOCOL)

def load_object_from_compressed_file(file_path, verbose = 1):
    with_pickle_suffix = file_path + '.pickle'
    if verbose == 1:
        print('loading object from compressed file:', with_pickle_suffix)
    with gzip.GzipFile(with_pickle_suffix, 'r') as f:
        obj = pickle.load(f)
    return obj

def save_object_as_compressed_file(obj, file_path):
    with_pickle_suffix = file_path + '.pickle'
    print('saving object to compressed file:', with_pickle_suffix)
    with gzip.GzipFile(with_pickle_suffix, 'w') as file:
        pickle.dump(obj, file, protocol=pickle.HIGHEST_PROTOCOL)

def get_num_files_in_directory(directory, prefix = None, suffix = None):
    count = 0
    for _, _, files in os.walk(directory):
        for file_name in files:
            if prefix is not None:
                if not file_name.startswith(prefix):
                    continue
            
            if suffix is not None:
                if not file_name.endswith(suffix):
                    continue
                
            count += 1
    return count

def write_text_to_file(text, full_path):
    text_file = open(f"{full_path}.txt", "w")
    text_file.write(text)
    text_file.close()

def read_all_text_from_file(full_path):
    file = open(full_path,mode='r')
    all_text = file.read()
    file.close() 
    return all_text

def loop_file_names_in_directory(root):
    for dir_path, dir_names, file_names in os.walk(root):
        for file_name in file_names:
            full_file_path = dir_path + os.sep + file_name
            yield full_file_path



from torch.optim.lr_scheduler import LambdaLR

class DoNothingToLR():
    def __init__(self, optimizer=None, last_epoch=-1):
        pass
    
    def step(self):
        pass

    def state_dict(self):
        return {}

    def load_state_dict(self, checkpoint):
        pass


class KerasDecay():
    def __init__(self, optimizer, decay, last_step=-1):
        self.decay = decay
        # THE KERAS DECAY FORMULA
        # LR = INIT_LR * (1.0/(1.0 + decay * step))
        self.fcn = lambda step: (1./(1. + self.decay * step)) 
        self.inner_scheduler = LambdaLR(optimizer, lr_lambda=self.fcn, last_epoch=last_step)

    def step(self):
        self.inner_scheduler.step()

    def state_dict(self):
        return self.inner_scheduler.state_dict()

    def load_state_dict(self, checkpoint):
        self.inner_scheduler.load_state_dict(checkpoint)




from lpd.enums.metric_method import MetricMethod
from lpd.metrics.metric_base import MetricConfusionMatrixBase
from lpd.metrics.confusion_matrix import ConfusionMatrix



class Stats():
    def __init__(self, metric_method: MetricMethod):
        self.sum = 0
        self.count = 0
        self.last = 0
        self.metric_method = metric_method
        self.reset()

    def reset(self):
        self.sum = 0
        self.count = 0
        self.last = 0

    def add_value(self, value, count):
        if self.metric_method == MetricMethod.MEAN:
            self.sum += value * count
            self.count += count

        elif self.metric_method == MetricMethod.SUM:
            self.sum += value
            self.count += count
        
        elif self.metric_method == MetricMethod.LAST:
            self.sum = value
            self.count = 1

        self.last = value

    def get_value(self):
        if self.metric_method == MetricMethod.MEAN:
            if self.count == 0:
                return 0
            return self.sum/self.count

        elif self.metric_method == MetricMethod.SUM:
            return self.sum

        elif self.metric_method == MetricMethod.LAST:
            return self.sum

class StatsResult():
    def __init__(self, trainer_name, stats):
        self.trainer_name = trainer_name
        self.loss = stats.get_loss().tolist()
        self.metrics = {name:value.tolist() for name,value in stats.get_metrics().items()}

    def __str__(self):
        metrics_str = 'Metrics: no metrics found'
        if self.metrics:
            metrics_str = f'Metrics: {self.metrics}'

        return '------------------\n' \
               'Evaluation Result:\n' \
               f'Trainer name: "{self.trainer_name}"\n' \
               f'Loss: {self.loss}\n' \
               f'{metrics_str}\n' \
               '------------------'

class TrainerStats():
    def __init__(self, metric_name_to_func):
        self.metric_name_to_func = metric_name_to_func
        self.loss_stats = Stats(MetricMethod.MEAN)
        self.metric_name_to_stats = {metric_name:Stats(metric.metric_method) for metric_name,metric in self.metric_name_to_func.items()}
        self.confusion_matrix = self._handle_confusion_matrix()

    def _handle_confusion_matrix(self):
        for _,metric in self.metric_name_to_func.items():
            if isinstance(metric, MetricConfusionMatrixBase):
                confusion_matrix = ConfusionMatrix(num_classes=metric.num_classes, 
                                                   labels=metric.labels, 
                                                   predictions_to_classes_convertor=metric.predictions_to_classes_convertor, 
                                                   threshold=metric.threshold)
                # NEED A SINGLE CONFUSION MATRIX PER TRAINER-STATS
                return confusion_matrix
        return None

    def reset(self):
        self.loss_stats.reset()

        for metric_name, stats in self.metric_name_to_stats.items():
            stats.reset()

        if self.confusion_matrix:
            self.confusion_matrix.reset()

    def add_loss(self, loss, batch_size):
        self.loss_stats.add_value(loss.clone().detach(), batch_size)

    def add_metrics(self, y_pred, y_true, batch_size):
        if self.confusion_matrix:
            self.confusion_matrix.update_state(y_pred, y_true)
            for _, metric in self.metric_name_to_func.items():
                if isinstance(metric, MetricConfusionMatrixBase):
                    metric._set_confusion_matrix(self.confusion_matrix)

        for metric_name, stats in self.metric_name_to_stats.items():
            metric_func = self.metric_name_to_func[metric_name]
            metric_value = metric_func(y_pred, y_true)
            stats.add_value(metric_value.clone().detach(), batch_size)

    def get_loss(self):
        return self.loss_stats.get_value()

    def get_metrics(self):
        return {metric_name:stats.get_value() for metric_name, stats in self.metric_name_to_stats.items()}


import torch as T
from lpd.trainer import Trainer
from lpd.enums import Phase, State

class Predictor():
    """
        slim class to make predictions on trained model

        Args:
            model - your model (nn.Module)
            device - the device to send the inputs to
            callbacks - (optional) list of lpd.callbacks to apply during the predictions phases and states
                        callbacks will be executed by the order of the list 
            name - a friendly identifier

        Methods:
            from_trainer - for creating a new Predictor instance from given Trainer
            from_checkpoint - for creating a new Predictor instance from a saved Trainer checkpoint
            predict_sample - make prediction on single sample
            predict_batch - make prediction on single batch
            predict_data_loader - make prediction on data loader (DataLoader/Iterable/Generator)
    """

    def __init__(self, model,
                       device,
                       callbacks=None,
                       name='lpd predictor'):
        self._inner = Trainer(model=model, 
                             device=device, 
                             loss_func=None,
                             optimizer=None,
                             scheduler=None,
                             metric_name_to_func=None,
                             train_data_loader=None,
                             val_data_loader=None,
                             train_steps=0,
                             val_steps=0,
                             callbacks=callbacks,
                             name=name)

    @staticmethod
    def from_trainer(trainer):
        print(f'[Predictor] - Loading from trainer {trainer.name}')

        predictor = Predictor(model=trainer.model, 
                              device=trainer.device, 
                              callbacks=trainer.callbacks,
                              name=f'Predictor-for-Trainer-{trainer.name}')

        return predictor

    @staticmethod
    def from_checkpoint(dir_path,
                        file_name,
                        model, 
                        device):
        full_path = dir_path + file_name
        checkpoint = T.load(full_path, map_location=device)
        print(f'[Predictor] - Loading from {full_path}')
        model.load_state_dict(checkpoint['model'])

        callbacks = None
        if 'callbacks' in checkpoint:
            callbacks = checkpoint['callbacks']

        name = None
        if 'name' in checkpoint:
            name = checkpoint['name']
            
        predictor = Predictor(model=model, 
                              device=device, 
                              callbacks=callbacks,
                              name=name)

        return predictor

    def predict_sample(self, inputs):
        return self._inner.predict_sample(inputs)

    def predict_batch(self, inputs):
        return self._inner.predict_batch(inputs)

    def predict_data_loader(self, inputs_data_loader, steps):
        return self._inner.predict_data_loader(inputs_data_loader, steps)






from lpd.trainer_stats import TrainerStats
from lpd.enums import Phase, State, MonitorType, MonitorMode, StatsType
from lpd.callbacks.callback_base import CallbackBase
from lpd.callbacks.callback_context import CallbackContext
from lpd.callbacks.callback_monitor import CallbackMonitorResult

from torch.utils.tensorboard import SummaryWriter
from typing import Union, List, Optional, Dict

class Tensorboard(CallbackBase):
    """ 
        Writes entries directly to event files in the summary_writer_dir to be
        consumed by TensorBoard.
        Args:
            apply_on_phase - see in CallbackBase
            apply_on_states - see in CallbackBase
            summary_writer_dir - the folder path to save tensorboard output
                                 if passed None, will write to the current dir
    """

    def __init__(self, apply_on_phase: Phase=Phase.EPOCH_END, 
                        apply_on_states: Union[State, List[State]]=State.EXTERNAL,
                        summary_writer_dir: str=None):
        super(Tensorboard, self).__init__(apply_on_phase, apply_on_states)
        self.TRAIN_NAME = 'Train'
        self.VAL_NAME = 'Val'
        self.summary_writer_dir = summary_writer_dir
        if self.summary_writer_dir is None:
            raise ValueError("[Tensorboard] - summary_writer_dir was not provided")
        self.tensorboard_writer = SummaryWriter(summary_writer_dir + 'tensorboard_files')

    def _write_to_summary(self, phase_name: str ,epoch: int, stats: TrainerStats):
        self.tensorboard_writer.add_scalar(f'{phase_name} loss', stats.get_loss(), global_step=epoch)
        for metric_name, value in stats.get_metrics().items():
            self.tensorboard_writer.add_scalar(f'{phase_name} {metric_name}', value, global_step=epoch)

    def __call__(self, callback_context: CallbackContext):
        c = callback_context #READABILITY DOWN THE ROAD
        self._write_to_summary(self.TRAIN_NAME, c.epoch, c.train_stats)
        self._write_to_summary(self.VAL_NAME, c.epoch, c.val_stats)
